/*
 * Copyright 2012 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.odftoolkit.odfdom.component;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.odftoolkit.odfdom.doc.OdfDocument;
import org.odftoolkit.odfdom.doc.table.OdfTable;
import org.odftoolkit.odfdom.doc.table.OdfTableRow;
import org.odftoolkit.odfdom.dom.OdfContentDom;
import org.odftoolkit.odfdom.dom.OdfDocumentNamespace;
import org.odftoolkit.odfdom.dom.OdfStylesDom;
import static org.odftoolkit.odfdom.component.JsonOperationProducer.*;
import org.odftoolkit.odfdom.dom.attribute.draw.DrawStyleNameAttribute;
import org.odftoolkit.odfdom.dom.attribute.table.TableStyleNameAttribute;
import org.odftoolkit.odfdom.dom.attribute.text.TextStyleNameAttribute;
import org.odftoolkit.odfdom.dom.element.OdfStylableElement;
import org.odftoolkit.odfdom.dom.element.OdfStyleBase;
import org.odftoolkit.odfdom.dom.element.OdfStylePropertiesBase;
import org.odftoolkit.odfdom.dom.element.draw.DrawFrameElement;
import org.odftoolkit.odfdom.dom.element.draw.DrawImageElement;
import org.odftoolkit.odfdom.dom.element.office.OfficeFontFaceDeclsElement;
import org.odftoolkit.odfdom.dom.element.style.StyleChartPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleDefaultStyleElement;
import org.odftoolkit.odfdom.dom.element.style.StyleDrawingPagePropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleFontFaceElement;
import org.odftoolkit.odfdom.dom.element.style.StyleGraphicPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleHeaderFooterPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleListLevelLabelAlignmentElement;
import org.odftoolkit.odfdom.dom.element.style.StyleListLevelPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StylePageLayoutPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleParagraphPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleRubyPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleSectionPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleStyleElement;
import org.odftoolkit.odfdom.dom.element.style.StyleTableCellPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleTableColumnPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleTablePropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleTableRowPropertiesElement;
import org.odftoolkit.odfdom.dom.element.style.StyleTextPropertiesElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableCellElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableColumnElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableColumnGroupElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableColumnsElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableHeaderColumnsElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableHeaderRowsElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableRowElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableRowGroupElement;
import org.odftoolkit.odfdom.dom.element.table.TableTableRowsElement;
import org.odftoolkit.odfdom.dom.element.text.TextAElement;
import org.odftoolkit.odfdom.dom.element.text.TextHElement;
import org.odftoolkit.odfdom.dom.element.text.TextLineBreakElement;
import org.odftoolkit.odfdom.dom.element.text.TextListElement;
import org.odftoolkit.odfdom.dom.element.text.TextListHeaderElement;
import org.odftoolkit.odfdom.dom.element.text.TextListItemElement;
import org.odftoolkit.odfdom.dom.element.text.TextListLevelStyleElementBase;
import org.odftoolkit.odfdom.dom.element.text.TextListLevelStyleImageElement;
import org.odftoolkit.odfdom.dom.element.text.TextPElement;
import org.odftoolkit.odfdom.dom.element.text.TextParagraphElementBase;
import org.odftoolkit.odfdom.dom.element.text.TextSpanElement;
import org.odftoolkit.odfdom.dom.element.text.TextTabElement;
import org.odftoolkit.odfdom.dom.element.text.TextTextInputElement;
import org.odftoolkit.odfdom.dom.style.OdfStyleFamily;
import org.odftoolkit.odfdom.dom.style.props.OdfStylePropertiesSet;
import org.odftoolkit.odfdom.incubator.doc.draw.OdfDrawImage;
import org.odftoolkit.odfdom.incubator.doc.office.OdfOfficeAutomaticStyles;
import org.odftoolkit.odfdom.incubator.doc.office.OdfOfficeStyles;
import org.odftoolkit.odfdom.incubator.doc.style.OdfDefaultStyle;
import org.odftoolkit.odfdom.incubator.doc.style.OdfStyle;
import org.odftoolkit.odfdom.incubator.doc.text.OdfTextListStyle;
import org.odftoolkit.odfdom.pkg.OdfElement;
import org.odftoolkit.odfdom.pkg.OdfFileDom;
import org.odftoolkit.odfdom.pkg.OdfPackage;
import org.odftoolkit.odfdom.pkg.manifest.OdfFileEntry;
import org.odftoolkit.odfdom.type.Base64Binary;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * ToDo: Is it more flexible to build a different queue for OperationQueue and
 * create an JSON exporter? Can a JSONArray / JSONObject be initialized with an
 * existing queue?
 *
 * @author svante.schubertATgmail.com
 */
//ToDo *** AFTER_JSON_SUPPORT ***   class JsonOperationImporter {
public class JsonOperationConsumer {

	private static final Logger LOG = Logger.getLogger(JsonOperationConsumer.class.getName());
	private static JSONObject CELL_WITH_BORDER_ATTRS = null;

	static {
		try {
			CELL_WITH_BORDER_ATTRS = new JSONObject(new JSONTokener("{\"cell\":{\"padding\":97,\"borderLeft\":{\"width\":2,\"style\":\"solid\",\"color\":{\"value\":\"000000\",\"type\":\"rgb\"}},\"borderBottom\":{\"width\":2,\"style\":\"solid\",\"color\":{\"value\":\"000000\",\"type\":\"rgb\"}},\"borderTop\":{\"width\":2,\"style\":\"solid\",\"color\":{\"value\":\"000000\",\"type\":\"rgb\"}},\"borderRight\":{\"width\":2,\"style\":\"solid\",\"color\":{\"value\":\"000000\",\"type\":\"rgb\"}}}}"));
		} catch (JSONException ex) {
			Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	/**
	 * For the demo document "FruitDepot-SeasonalFruits.odt" the table looks
	 * like <text:p text:style-name="Text_20_body"/>
	 *
	 * @param rootComponent high level document structure
	 * @param paraPos position of the paragraph starting with 0 and one over the
	 * existing is allowed
	 * @throws IndexOutOfBoundsException - if index is out of range (index < 0
	 * || index > size()). One over size is allowed to append a paragraph.
	 */
	public static void insertParagraph(Component rootComponent, JSONArray start, JSONObject attrs) throws IndexOutOfBoundsException {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);

		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			// CREATING NEW ROOT ELEMENT		
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			//If there are any list properties there have to be a list inserted, check if preceding- & following-sibling is a list?
			TextParagraphElementBase paragraphBaseElement;
			JSONObject paraProps = null;
			if (attrs != null) {
				paraProps = attrs.optJSONObject("paragraph");
				if (paraProps != null) {
					int outlineLevel = paraProps.optInt("outlineLevel");
					if (outlineLevel != 0) {
						paragraphBaseElement = new TextHElement(xmlDoc);
						((TextHElement) paragraphBaseElement).setTextOutlineLevelAttribute(outlineLevel + 1);
					} else {
						paragraphBaseElement = new TextPElement(xmlDoc);
					}
				} else {
					paragraphBaseElement = new TextPElement(xmlDoc);
				}
			} else {
				paragraphBaseElement = new TextPElement(xmlDoc);
			}

			try {
				// ADDING COMPONENT
				addComponents(parentComponent, paragraphBaseElement, start.getInt(start.length() - 1), 1);
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
			// ADDING STYLES TO THE NEW ELEMENT
			StyleStyleElement autoStyle = addStyle(attrs, paragraphBaseElement, xmlDoc);
			if (paraProps != null) {
				if (paraProps.has("listLevel")) {
					if (paraProps.has("listStyleId")) {
						String listStyleId = paraProps.optString("listStyleId");
						if (listStyleId != null && !listStyleId.isEmpty()) {
							autoStyle.setStyleListStyleNameAttribute(listStyleId);
						}
					}
					int listLevel = getListLevel(paragraphBaseElement);
					int newListLevel = paraProps.optInt("listLevel", -2);
					// if the list level should be unchanged
					if (newListLevel == -2) {
						// use the current listlevel..
						newListLevel = listLevel;
					}
					// we need to keep the existing top level list attributes, at least the list style 
					String currentListStyleName = null;
					TextListElement rootListElement = getListRootElement(paragraphBaseElement);
					if (rootListElement != null) {
						currentListStyleName = rootListElement.getTextStyleNameAttribute();
					}
					setParagraphListProperties(paragraphBaseElement, paraProps, xmlDoc, listLevel, newListLevel, currentListStyleName);
				}
			}
		}
	}

	public static void insertTab(Component rootComponent, JSONArray start, JSONObject attrs) throws IndexOutOfBoundsException {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);

		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			// CREATING NEW ROOT ELEMENT		
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			OdfElement newElement = new TextTabElement(xmlDoc);
			// If there are hard coded styles on the line break, add them to a text:span element surrounding the text:line-break
			// At all places of a text:tab a text:span as potential parent is allowed:
			// http://docs.oasis-open.org/office/v1.2/os/OpenDocument-v1.2-os-part1.html#element-text_tab
			if (attrs != null) {
				TextSpanElement newSpanElement = new TextSpanElement(xmlDoc);

				// ADDING STYLES TO THE NEW ELEMENT
				addStyle(attrs, newSpanElement, xmlDoc);
				newSpanElement.appendChild(newElement);
				newElement = newSpanElement;
			}
			try {
				// ADDING COMPONENT
				addComponents(parentComponent, newElement, start.getInt(start.length() - 1), 1);
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
	}

	public static void insertHardBreak(Component rootComponent, JSONArray start, JSONObject attrs) throws IndexOutOfBoundsException {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);

		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			// CREATING NEW ROOT ELEMENT		
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			OdfElement newElement = new TextLineBreakElement(xmlDoc);
			// If there are hard coded styles on the line break, add them to a text:span element surrounding the text:line-break
			// At all places of a text:line-break a text:span as potential parent is allowed:
			// http://docs.oasis-open.org/office/v1.2/os/OpenDocument-v1.2-os-part1.html#element-text_line-break
			if (attrs != null) {
				TextSpanElement newSpanElement = new TextSpanElement(xmlDoc);

				// ADDING STYLES TO THE NEW ELEMENT
				addStyle(attrs, newSpanElement, xmlDoc);
				newSpanElement.appendChild(newElement);
				newElement = newSpanElement;
			}
			try {
				// ADDING COMPONENT
				addComponents(parentComponent, newElement, start.getInt(start.length() - 1), 1);
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
	}

	/**
	 * @return the name of the style, which is currently is hidden in the main
	 * property collections
	 */
	private static String getTemplateStyleName(JSONObject attrs) {
		String styleName = null;
		if (attrs != null && attrs.has("styleId")) {
			try {
				styleName = attrs.getString("styleId");
				if (styleName == null) {
					styleName = null;
				}
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
		return styleName;
	}

	private static StyleStyleElement addStyle(JSONObject attrs, OdfStylableElement newElement, OdfFileDom ownerDocument) {
		StyleStyleElement autoStyle = null;
		// temporary font properties taken into adapter
		OdfDocument doc4Fonts = ((OdfDocument) ownerDocument.getDocument());

		if (attrs != null) {
			OdfStyleFamily styleFamily = newElement.getStyleFamily();

			String styleName = getTemplateStyleName(attrs);
			if (styleName != null && !styleName.isEmpty()) {
//				Map<String, OdfStylePropertiesSet> elementProps = Component.getAllOxStyleGroupingIdProperties(styleFamily);
//				JSONObject autoProps = null;

//				boolean hasOnlyTemplateStyle = true;
//				for (String paraPropsId : elementProps.keySet()) {
//					autoProps = attrs.optJSONObject(paraPropsId);
//					if (paraPropsId.equals("paragraph")) {
//						if (!(autoProps == null || autoProps.equals(null)
//								|| autoProps != null
//								&& (attrs.length() == 1 && autoProps.length() == 1
//								|| attrs.length() == 1 && autoProps.length() == 2 && autoProps.has("outlineLevel")))) {
//							hasOnlyTemplateStyle = false;
//							break;
//						}
//// ToDo: Improvement: EVERY ELEMENT PROPERTIES HAVE TO BE SUBSTRACTED TO NOT CREATE EMPTY STYLES! E.g. counting off those properties..				
////					} if (paraPropsId.equals("drawing")) {
////						if (!(autoProps == null || autoProps.equals(null) ||
////								autoProps != null && 
////								(attrs.length() == 1 && autoProps.length() == 1 || 
////								attrs.length() == 1 && autoProps.length() == 2 && autoProps.has("outlineLevel")))) {
////							hasOnlyTemplateStyle = false;
////							break;
////						}
//					} else {
//						if (autoProps != null && autoProps.length() != 0) {
//							hasOnlyTemplateStyle = false;
//							break;
//						}
//					}
//				}

				// IF ONLY TEMPLATE STYLE
				if (!hasHardProperties(attrs, styleFamily)) {
					// Add template style, if there was no hard formatting it will be set directly, otherwise the template style has to reference					
					addStyleNameAttribute(newElement, styleFamily, (OdfFileDom) ownerDocument, styleName);

					// IF THERE ARE HARD FORMATTING STYLES AND TEMPLATE STYLES
				} else {
					autoStyle = newElement.getOrCreateUnqiueAutomaticStyle();
					addStyleNameAttribute(newElement, styleFamily, (OdfFileDom) ownerDocument, autoStyle.getStyleNameAttribute());
					autoStyle.setStyleParentStyleNameAttribute(styleName);

					// APPLY HARD FORMATTING PROPERTIES OF ODF FAMILY
					mapProperties(styleFamily, attrs, autoStyle, doc4Fonts);
				}
				// IF ONLY AUTOMATIC STYLE	
			} else {
				autoStyle = newElement.getOrCreateUnqiueAutomaticStyle();
				addStyleNameAttribute(newElement, styleFamily, (OdfFileDom) ownerDocument, autoStyle.getStyleNameAttribute());

				// APPLY HARD FORMATTING PROPERTIES OF ODF FAMILY
				mapProperties(styleFamily, attrs, autoStyle, doc4Fonts);
			}
		}
		return autoStyle;
	}

	private static void addStyleNameAttribute(OdfElement newElement, OdfStyleFamily styleFamily, OdfFileDom ownerDocument, String styleName) {
		// Add template style, if there was no hard formatting it will be set directly, otherwise the template style has to reference
		if (styleFamily.getName().equals("paragraph") || styleFamily.getName().equals("text")) {
			TextStyleNameAttribute attr = new TextStyleNameAttribute(ownerDocument);
			newElement.setOdfAttribute(attr);
			attr.setValue(styleName);
		} else if (styleFamily.getName().equals("table") || styleFamily.getName().equals("table-column") || styleFamily.getName().equals("table-row") || styleFamily.getName().equals("table-cell")) {
			TableStyleNameAttribute attr = new TableStyleNameAttribute(ownerDocument);
			newElement.setOdfAttribute(attr);
			attr.setValue(styleName);
		} else if (styleFamily.getName().equals("graphic")) {
			DrawStyleNameAttribute attr = new DrawStyleNameAttribute(ownerDocument);
			newElement.setOdfAttribute(attr);
			attr.setValue(styleName);
		}
	}

	/**
	 * Move the pointed component to its new destination.
	 *
	 * @param from the origin of the component to be moved
	 * @param the the new destination
	 */
	public static void moveDrawing(Component rootComponent, JSONArray start, JSONArray to) {
		Component parentSourceComponent = rootComponent.getParentOf(start);
		OdfElement parentSource = (OdfElement) parentSourceComponent.getRootElement();
		Component parentDestinationComponent = rootComponent.getParentOf(to);
		OdfElement parentDestination = (OdfElement) parentDestinationComponent.getRootElement();

		// MOVING COMPONENT - ONLY DRAWING (TEXT CONTENT) SUPPORTED
		try {
			Node movedNode = parentSource.receiveNode(start.getInt(start.length() - 1));
			if (movedNode != null && movedNode instanceof OdfElement) {
				OdfElement movedElement = (OdfElement) movedNode;
				movedElement = (OdfElement) parentSource.removeChild(movedElement);
				if (movedElement != null) {
					parentDestination.insert(movedElement, to.getInt(to.length() - 1));
				} else {
					LOG.severe("Could not find component at position: " + start);
				}
			} else {
				LOG.severe("Could not find component at position: " + start + ", only: " + movedNode);
			}
		} catch (JSONException ex) {
			Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	/**
	 * Deletes a component from the document.
	 *
	 * @param rootComponent the root of the component tree
	 * @param start the position of the component to be deleted /* ToDo:
	 * Possible to join the two delete methods, this and OdfElement delete(int
	 * start, int end)??
	 */
	public static void delete(Component rootComponent, JSONArray start) throws IndexOutOfBoundsException {
		Component parentComponent = rootComponent.getParentOf(start);
		OdfElement parent = (OdfElement) parentComponent.getRootElement();


		OdfElement targetElement = null;
		// DELETING TEXT
		if (parentComponent instanceof TextComponent) {
			int pos = start.optInt(start.length() - 1);
			((TextComponent) parentComponent).removeText(pos, pos + 1);
		} else {
			// DELETING COMPONENT
			try {
				targetElement = (OdfElement) parentComponent.getChild(start.getInt(start.length() - 1)).getRootElement();
				if (targetElement != null) {
					OdfElement targetParent = (OdfElement) targetElement.getParentNode();
					if (targetParent.equals(parent)) {
						parent.removeChild(targetElement);
					} else {
						// if the parent component root element is not the parent of the child component, there have to be boilerplate elements inbetween that have to take care of
						OdfElement.removeComponentElementAndEmptyBoilerplate(parent, targetElement);
					}
					try {
						parentComponent.remove(start.getInt(start.length() - 1));
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else {
					LOG.severe("Could not find component at position: " + start);
				}
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
	}

	//ToDo *** AFTER_JSON_SUPPORT ***   private static void insertParagraph(ComponentTree rootComponent, int pos){		
	public static void insertText(Component rootComponent, JSONArray start, JSONObject attrs, String newText) throws IndexOutOfBoundsException {
		Component parentComponent = rootComponent.getParentOf(start);
		if (parentComponent != null) {
			TextParagraphElementBase componentRootElement = (TextParagraphElementBase) parentComponent.getRootElement();

			try {
				// ToDo - parse the text spans positions from the beginning, creating array - could do this from the beginning..
				// better not for very long texts it is unnecessary, does not scale
				//        adding the span depths, the span element referece, for every character?			
				componentRootElement.insert(newText, start.getInt(start.length() - 1));
				if (attrs != null && attrs.length() > 0) {
					if (hasHardProperties(attrs, OdfStyleFamily.Text)) {
						try {
							// CREATING NEW ROOT ELEMENT		
							OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
							TextSpanElement newSpanElement = new TextSpanElement(xmlDoc);
							// Map the given attributes (i.e. attrs) to ODF attributes to make span style inheritance possible
							// ADDING STYLES TO THE NEW ELEMENT
							addStyle(attrs, newSpanElement, xmlDoc);
							int endPos = start.getInt(start.length() - 1) + newText.length() - 1;
							componentRootElement.markText(start.getInt(start.length() - 1), endPos, newSpanElement);
						} catch (JSONException ex) {
							Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
						}
					}
				}
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		} else {
			LOG.log(Level.SEVERE, "Could not insertText as no container (e.g. paragraph) was found at position: {0}", start);
		}
	}

	private static boolean hasHardProperties(JSONObject attrs, OdfStyleFamily styleFamily) {
		int attrsLength = attrs.length();
		if (attrs.has("styleId") && attrs.isNull("styleId")) {
			attrsLength--;
		}
		Map<String, OdfStylePropertiesSet> elementProps = Component.getAllOxStyleGroupingIdProperties(styleFamily);
		for (String propertyId : elementProps.keySet()) {
			if (attrs.has(propertyId)) {
				JSONObject newProps = attrs.optJSONObject(propertyId);
				int propsLength = newProps.length();
				Iterator<String> keys = newProps.keys();
				String key = null;
				while (keys.hasNext()) {
					key = keys.next();
					if (newProps.has(key) && newProps.isNull(key)) {
						propsLength--;
					} else {
						break;
					}
				}
				if (propsLength == 0) {
					attrsLength--;
				} else {
					break;
				}
			}
		}
		return attrsLength != 0;
	}

	private static void mergeParagraph(Component rootComponent, JSONArray start) throws IndexOutOfBoundsException {
		Component firstComponent = rootComponent.get(start);
		TextParagraphElementBase firstParagraph = null;
		if (firstComponent != null) {
			firstParagraph = (TextParagraphElementBase) firstComponent.getRootElement();
			if (firstParagraph == null) {
				throw new IndexOutOfBoundsException("There was no component for " + start + " accessable.");
			}
		} else {
			throw new IndexOutOfBoundsException("There was no component for " + start + " accessable.");
		}

		Component secondComponent = rootComponent.getNextSiblingOf(start);
		if (secondComponent != null) {
			TextParagraphElementBase secondParagraph = (TextParagraphElementBase) secondComponent.getRootElement();
			if (secondParagraph != null) {
				try {
					secondParagraph.moveChildrenTo(firstParagraph);
					secondParagraph.getParentNode().removeChild(secondParagraph);
					secondComponent.getParent().remove(start.getInt(start.length() - 1) + 1);
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
				}
			} else {
				throw new IndexOutOfBoundsException("There was no sibling for " + start + " accessable.");
			}
		} else {
			LOG.severe("Could not fine second Paragraph to merge. Position: " + start);
		}
	}

	//ToDo *** AFTER_JSON_SUPPORT ***   private static void insertParagraph(ComponentTree rootComponent, int pos){		
	public static void delete(Component rootComponent, JSONArray start, JSONArray end) throws IndexOutOfBoundsException {
		// There is only an end position for text deletion
		if (end != null && !end.equals(start)) {
			// Delete the text in the first paragraph till the end
			deleteText(rootComponent, start, end); // "0-N"
		} else { // the paragraphs are the same, only text within a single paragraph
			delete(rootComponent, start);
		}
		// Split the single command depending on para pos to:
		// 1) Start  Paragraph: An inner text delete (delete after position)
		// 2) Middle Paragraph(s): paragraph delte range to be deleted 
		// 3) End    Paragraph: An inner text delete (delete before position)		
		// If not the same para, it is more than one command..
//        if (paraPosStart != paraPosEnd) {
//            // Delete the text in the first paragraph till the end
//            delete(rootComponent, paraPosStart, textPosStart, Integer.MAX_VALUE); // "0-N"
//            // Delete the text in the last paragraph till the desired position
//            delete(rootComponent, paraPosEnd, 0, textPosEnd); // talic-underline<text:span text:style-name="T4">51-Bold-underline</text:span>68-Bold-italic-underline2</text:span>93-Bold</text:span>100-Normal!!
//            mergeParagraph(rootComponent, paraPosStart, paraPosEnd);
//
//            // FIXME: As soon we allow hierachies components might be in any arbitrary horizontal level
//            if ((paraPosStart + 1) != paraPosEnd) {
//                // remove all paragraph that are inbetween the two paras and the final one as it had been merged
//                for (int i = paraPosStart + 1; i <= paraPosEnd; i++) {
//                    // delete paragraph
//                    deleteParagraph(rootComponent, i);
//                }
//            }
//        } else { // the paragraphs are the same, only text within a single paragraph
//            delete(rootComponent, paraPosStart, textPosStart, textPosEnd);
//        }
	}

	/**
	 * Removes a template style from the document
	 *
	 * @param doc4Fonts temporary font properties taken into adapter
	 */
	public static void deleteStyleSheet(OdfDocument doc4Fonts, String styleId, String type) {
		OdfOfficeStyles styles = doc4Fonts.getDocumentStyles();
		OdfStyle oldStyle = styles.getStyle(styleId, Component.getFamily(type));
		styles.removeChild(oldStyle);
	}

//	/**
//	 * Adds a new none automatic style to the document. If the style is hidden a
//	 * default style, otherwise a template style will be added.
//	 * @param doc4Fonts  temporary font properties taken into adapter		
//	 */
	public static void insertStyleSheet(OdfDocument doc4Fonts, String type, String styleId, String styleName, JSONObject attrs, String parent, Boolean hidden) {

		OdfOfficeStyles styles = doc4Fonts.getDocumentStyles();
		if (attrs != null) {
			OdfStyleFamily styleFamily = Component.getFamily(type);
			if (hidden) { // DEFAULT STYLE
				// Removing any existing default style
				OdfDefaultStyle oldDefaultStyle = styles.getDefaultStyle(styleFamily);
				if (oldDefaultStyle != null) {
					styles.removeChild(oldDefaultStyle);
				}
				// Adding new default style
				StyleDefaultStyleElement defaultStyleElement = styles.newStyleDefaultStyleElement(styleFamily.getName());
				mapProperties(styleFamily, attrs, defaultStyleElement, doc4Fonts);
			} else { // TEMPLATE STYLE									
				OdfStyle oldStyle = styles.getStyle(styleName, styleFamily);
				if (oldStyle != null) {
					styles.removeChild(oldStyle);
				} else {
					/* Workaround for OOo applicatoins
					 * Applications from the OpenOffice family are exchanging any space in their names with an '_20_' string. 
					 * Unfortunately they are using them equivalent in general, therefore the names 'Heading_20_2' and 'Heading 2' would result and into a name clash */
					if (styleName != null && styleName.contains(" ")) {
						String testName = styleName.replace(" ", "_20_");
						oldStyle = styles.getStyle(testName, styleFamily);
						if (oldStyle != null) {
							styles.removeChild(oldStyle);
						}
					}
				}
				StyleStyleElement style = styles.newStyle(styleId, styleFamily);
				if (parent != null) {
					style.setStyleParentStyleNameAttribute(parent);
				}
				if (styleName != null && !styleName.isEmpty()) {
					style.setStyleDisplayNameAttribute(styleName);
				}

				// APPLY FORMATTING PROPERTIES 
				if (type.equals("table")) {
					mapProperties(styleFamily, attrs.optJSONObject("wholetable"), style, doc4Fonts);
				} else {
					mapProperties(styleFamily, attrs, style, doc4Fonts);
				}


				// APPLY PARAGRAPH ONLY PROPERTIES 
				if (styleFamily.equals(OdfStyleFamily.Paragraph)) {
					JSONObject props = attrs.optJSONObject("paragraph");
					if (props != null) {
						// Heading Level
						if (props.has("outlineLevel")) {
							if (props.hasAndNotNull("outlineLevel")) {
								Integer outlineLevel = props.optInt("outlineLevel");
								if (outlineLevel > -1) {
									style.setStyleDefaultOutlineLevelAttribute(outlineLevel + 1);
								} else {
									style.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "default-outline-level");
								}
							} else {
								style.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "default-outline-level");
							}
						}

						// Follow-Up Style
						String nextStyleId = props.optString("nextStyleId");
						// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
						if (nextStyleId != null && !nextStyleId.isEmpty()) {
							style.setStyleNextStyleNameAttribute(nextStyleId);
						}
					}
				}
			}
		}
	}

	/**
	 * Setting attributes on text, require the text container, otherwise the
	 * character itself (as paragraph is not child of paragraph
	 */
	public static void setAttributes(Component rootComponent, JSONArray start, JSONArray end, JSONObject attrs) {
		// Only process the method if all mandatory values are present
		if (rootComponent != null && start != null && attrs != null & attrs.length() != 0) {
			// Parent will not change and have to exist to insert the new component				
			final Component parentComponent = rootComponent.getParentOf(start);
			if (parentComponent != null) {
				// the startComponent migth be either text or an element
				final Component startComponent = parentComponent.getChild(start.optInt(start.length() - 1));
				OdfElement targetElement;
				if (startComponent != null) {
					Component targetComponent = parentComponent.getChild(start.optInt(start.length() - 1));
					targetElement = targetComponent.getRootElement();
					if (targetElement != null & targetElement instanceof OdfStylableElement) {

						// ADDING STYLES TO THE NEW ELEMENT
						OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
						// Adds the automatic/tempate style to the target element
						StyleStyleElement autoStyle = addStyle(attrs, (OdfStylableElement) targetElement, xmlDoc);

						if (targetElement instanceof TextParagraphElementBase) {
							TextParagraphElementBase paragraphBaseElement = (TextParagraphElementBase) targetElement;
							// Check if the paragraph should become a list 
							JSONObject paraProps = attrs.optJSONObject("paragraph");
							if (paraProps != null) {
								// if there the paragraph list properties are changed..
								if (paraProps.has("listLevel") || paraProps.hasAndNotNull("listStyleId")) {
									// get the current list level
									int listLevel = getListLevel(paragraphBaseElement);
									// set listLevel to -2 if not contained
									int newListLevel = paraProps.optInt("listLevel", -2);
									// if the list level should be unchanged
									if (newListLevel == -2) {
										// use the current listlevel..
										newListLevel = listLevel;
									}

									// if the paragraph in the list should be shown..
									boolean listLabelHidden = false;
									if (paraProps.has("listLabelHidden") || paraProps.hasAndNotNull("listLabelHidden")) {
										// get the current list level
										listLabelHidden = paraProps.optBoolean("listLabelHidden");
									}

									// we need to keep the existing top level list attributes, at least the list style 
									String currentListStyleName = null;
									TextListElement rootListElement = getListRootElement(paragraphBaseElement);
									if (rootListElement != null) {
										currentListStyleName = rootListElement.getTextStyleNameAttribute();
									}

									// we need to split the paragraph first to root level, to be certain that it is first in a list-item
									if (paraProps.hasAndNotNull("listStyleId") && !listLabelHidden) {
										String listStyleId = paraProps.optString("listStyleId");
//if (7 == (start.optInt(start.length() - 1)) || listStyleId == "L4") {
//	System.out.println("x");
//}


										if (listStyleId != null && !listStyleId.isEmpty()) {
											autoStyle.setStyleListStyleNameAttribute(listStyleId);
										}
										// ToDo Optimiszation: remove the following for list level 0 and those without preceding/following content!
										// temporary move the paragraph out of the list to separate both list styles
										setParagraphListProperties(paragraphBaseElement, paraProps, xmlDoc, listLevel, -1, null);
										listLevel = -1;
									}
									setParagraphListProperties(paragraphBaseElement, paraProps, xmlDoc, listLevel, newListLevel, currentListStyleName);
								}
								// switch element if outlineLevel is on a paragraph (or vise versa)
								if (paraProps.has("outlineLevel") || attrs.has("styleId")) {
									int outlineLevel = -2;
									if (paraProps.hasAndNotNull("outlineLevel")) {
										outlineLevel = paraProps.optInt("outlineLevel");
										// if not the automatic/hard style is overwriting the template/referenced style
									} else if (attrs.has("styleId")) {
										if (attrs.hasAndNotNull("styleId")) {
											String styleId = attrs.optString("styleId");
											if (styleId != null && !styleId.isEmpty()) {
												outlineLevel = getDefaultOutlineLevelFromStyleHierarchy(paragraphBaseElement);
											}
										}
									}
									ensureCorrectParagraphElement(outlineLevel, (TextParagraphElementBase) targetElement, targetComponent);
								}
							}
						} else if (targetElement instanceof TableTableElement) {
							TableTableElement tableElement = (TableTableElement) targetElement;
							JSONObject tableProps = attrs.optJSONObject("table");
							if (tableProps != null) {
								JSONArray tableGrid = tableProps.optJSONArray("tableGrid");
								if (tableGrid != null) {
									setAbsoluteColumnWidths(tableElement, tableGrid);
								}
							}
						} else if (targetElement instanceof DrawFrameElement) {
							DrawFrameElement frameElement = (DrawFrameElement) targetElement;
							JSONObject drawingProps = attrs.optJSONObject("drawing");
							if (drawingProps != null) {
								setFrameProperties(frameElement, drawingProps);
								NodeList images = frameElement.getElementsByTagName("draw:image");
								if (images.getLength() > 0) {
									DrawImageElement imageElement = (DrawImageElement) images.item(0);
									setImageProperties(imageElement, drawingProps);
								}
							}
						}
					}
				}

				// If the parent is a paragraph/heading, the target can not be a paragraph/heading as they are not allwed nested.
				// Therefore we add a span around the text (or target element as draw:frame for image, which is done by MSO15 by default)
				OdfElement parentElement = parentComponent.getRootElement();
				if (parentElement != null && parentElement instanceof TextParagraphElementBase && attrs.has("character")) {
					TextParagraphElementBase destinationComponent = (TextParagraphElementBase) parentComponent.getRootElement();
					if (destinationComponent != null) {

						// ToDo - add the paragraph text length to the paragraph? Than I need to intercept all text manipulations!!
						// ToDo - parse the text spans positions from the beginning, creating array - could do this from the beginning..
						// better not for very long texts it is unnecessary, does not scale
						//  adding the span depths, the span element referece, for every character?			
						try {
							int startPos = start.getInt(start.length() - 1);
							int endPos;
							if (end == null) {
								endPos = startPos;
							} else {
								endPos = end.getInt(end.length() - 1);
							}
							String styleId = "";
							if (attrs.has("styleId")) {
								if (attrs.hasAndNotNull("styleId")) {
									styleId = (String) attrs.get("styleId");
								} else {
									styleId = null;
								}
							}

							// Insert an hyperlink above the span if an URL was included as property
							if (attrs.has("character")) {
								JSONObject charJSONProps = attrs.optJSONObject("character");
								String url = "";
								if (charJSONProps.has("url")) {
									if (charJSONProps.hasAndNotNull("url")) {
										url = (String) charJSONProps.get("url");
									} else {
										url = null;
									}
								}
								// CREATING NEW ROOT ELEMENT		
								OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
								TextSpanElement newSpanElement = new TextSpanElement(xmlDoc);
								// ADDING CHARACTER STYLES TO THE ADDITIONAL SPAN ELEMENT (NOTE: Important add style before marking the text..)
								addStyle(attrs, newSpanElement, xmlDoc);
								destinationComponent.markText(start.getInt(start.length() - 1), endPos, newSpanElement);

								if (url != null && !url.isEmpty() && !url.equals("null")) {
									TextAElement anchor = new TextAElement(xmlDoc);
									anchor.setXlinkHrefAttribute(url);
									anchor.setXlinkTypeAttribute("simple");

									// remove all text:a descendants
									NodeList anchors = newSpanElement.getElementsByTagNameNS(OdfDocumentNamespace.TEXT.getUri(), "a");
									int anchorCount = anchors.getLength();
									for (int i = anchorCount - 1; i >= 0; i--) {
										TextAElement a = (TextAElement) anchors.item(i);
										OdfElement.removeElement(a);
									}
									Element parent = (Element) newSpanElement.getParentNode();
									if (parent != null) {
										newSpanElement = (TextSpanElement) parent.replaceChild(anchor, newSpanElement);
										anchor.appendChild(newSpanElement);
									} else {
										destinationComponent.markText(start.getInt(start.length() - 1), endPos, anchor);
									}
								}
								// if any hyperlink should be removed!
								if (url == null) {
									// remove all text:a descendants
									NodeList anchors = newSpanElement.getElementsByTagNameNS(OdfDocumentNamespace.TEXT.getUri(), "a");
									int anchorCount = anchors.getLength();
									for (int i = anchorCount - 1; i >= 0; i--) {
										TextAElement a = (TextAElement) anchors.item(i);
										OdfElement.removeElement(a);
									}
								}
								// if some style is being removed, we need to adjust all descendants <text:span> and remove them if they are empty
								if (hasNullStyleProperty(charJSONProps) || (styleId == null)) {

									// REMOVE STYLES FOR ALL CHILDREN
									// get all span elements within the newSpan and remove the given styles/styleId	
									NodeList innerSpans = newSpanElement.getElementsByTagNameNS(OdfDocumentNamespace.TEXT.getUri(), "span");
									int spanCount = innerSpans.getLength();
									for (int i = spanCount - 1; i >= 0; i--) {
										TextSpanElement innerSpan = (TextSpanElement) innerSpans.item(i);
										OdfStyle autoStyle = (OdfStyle) addStyle(attrs, (OdfStylableElement) innerSpan, xmlDoc);
										String parentStyleName = null;
										if (autoStyle != null) {
											parentStyleName = ((StyleStyleElement) autoStyle).getStyleParentStyleNameAttribute();
										}
										if ((autoStyle == null || !autoStyle.hasPropertyAttribute() && parentStyleName == null || parentStyleName.isEmpty())) {
											OdfElement.removeElement(innerSpan);
										}
									}
									// REMOVE STYLES FOR ITSELF
									OdfStyle autoStyle = (OdfStyle) addStyle(attrs, (OdfStylableElement) newSpanElement, xmlDoc);
									String parentStyleName = null;
									if (autoStyle != null) {
										parentStyleName = ((StyleStyleElement) autoStyle).getStyleParentStyleNameAttribute();
									}
									if ((autoStyle == null || !autoStyle.hasPropertyAttribute()) && (parentStyleName == null || parentStyleName.isEmpty())) {
										OdfElement.removeElement(newSpanElement);
									} else // if hard styles were removed (value == null) and template style was removed (styleID == null)
									if (!hasNoneNullStyleProperty(charJSONProps) && (styleId == null || styleId.isEmpty())) {
										// remove the new added span, used only as temporaray delimiter									
										OdfElement.removeElement(newSpanElement);
									}
								}
							}
						} catch (JSONException ex) {
							Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
						}
					}
				}
			} else {
				LOG.log(Level.SEVERE, "No parent component found to ''setAttributes'' from {0} to {1} trying to add {2}", new Object[]{start, end, attrs});
			}
		} else {
			LOG.log(Level.SEVERE, "No parent component found to ''setAttributes'' from {0} to {1} trying to add {2}", new Object[]{start, end, attrs});
		}
	}

	/**
	 * if one property was set to NULL it has to be removed from the descendants
	 */
	private static boolean hasNullStyleProperty(JSONObject propertyContainer) {
		boolean hasNullProperty = false;
		Set<Entry<String, Object>> entrySet = propertyContainer.entrySet();
		for (Entry<String, Object> entry : entrySet) {
			Object value = entry.getValue();
			String key = entry.getKey();
			if ((value == null || value.equals(JSONObject.NULL) && !key.equals("url"))) {
				hasNullProperty = true;
				break;
			}
		}
		return hasNullProperty;
	}

	private static boolean hasNoneNullStyleProperty(JSONObject propertyContainer) {
		boolean hasProperty = false;
		Set<Entry<String, Object>> entrySet = propertyContainer.entrySet();
		for (Entry<String, Object> entry : entrySet) {
			Object value = entry.getValue();
			String key = entry.getKey();
			if ((value != null && !value.equals(JSONObject.NULL) && !key.equals("url"))) {
				hasProperty = true;
				break;
			}
		}
		return hasProperty;
	}

	private static int getDefaultOutlineLevelFromStyleHierarchy(TextParagraphElementBase paragraphBaseElement) {
		OdfStyle style = paragraphBaseElement.getDocumentStyle();
		int outlineLevel = getStyleOutlineLevel(style);

		if (style != null) {
			// if no outline Level was set, but there is a parent												
			if (outlineLevel == -2) {
				OdfStyleBase parentStyle = null;
				style.getParentStyle();
				while (outlineLevel == -2) {
					outlineLevel = getStyleOutlineLevel(parentStyle);
					if (outlineLevel != -2) {
						break;
					}
					if (parentStyle instanceof OdfStyle) {
						parentStyle = ((OdfStyle) parentStyle).getParentStyle();
					} else {
						break;
					}
					if (parentStyle == null) {
						break;
					}
				}
			}
		}
		return outlineLevel;
	}

	private static int getStyleOutlineLevel(OdfStyleBase style) {
		int outlineLevel = -2;
		if (style != null) {
			String outlineLevelValue = null;
			if (style.hasAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "default-outline-level")) {
				outlineLevelValue = style.getAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "default-outline-level");
				if (outlineLevelValue != null && !outlineLevelValue.isEmpty()) {
					outlineLevel = Integer.parseInt(outlineLevelValue);
				}
			}
		}
		return outlineLevel;
	}

	/**
	 * Exchanges a <text:p> element to a <text:h> when the paragraph receives an
	 * outline level
	 */
	private static void ensureCorrectParagraphElement(int outlineLevel, TextParagraphElementBase sourceElement, Component targetComponent) {
		TextParagraphElementBase newElement = null;
		if (outlineLevel > -1) {
			// switch to a heading
			if (!(sourceElement instanceof TextHElement)) {
				newElement = new TextHElement((OdfFileDom) sourceElement.getOwnerDocument());
				// add the outline level to the element				
				((TextHElement) newElement).setTextOutlineLevelAttribute(outlineLevel);
			} else {
				((TextHElement) sourceElement).setTextOutlineLevelAttribute(outlineLevel);
			}

			//} else if (outlineLevel == -1) {
			// The above should work, but is not supported by Frontend, issue?
		} else {
			// switch to a paragraph
			if (!(sourceElement instanceof TextPElement)) {
				newElement = new TextPElement((OdfFileDom) sourceElement.getOwnerDocument());
				// remove the outline level to the element				
			}
			sourceElement.removeAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "outline-level");
		}
		// if the type have to be changed, clone everything from the old to the new element
		if (newElement != null) {
			OdfElement parent = (OdfElement) sourceElement.getParentNode();
			parent.replaceChild(newElement, sourceElement);
			OdfElement.cloneNode(sourceElement, newElement, true);
			targetComponent.setRootElement(newElement);
			// adding back reference from element to component
			((OdfElement) newElement).setComponent(targetComponent);
		}
	}

	/**
	 * Deletes all hyperlink and template styles from any span/anchor in this
	 * area * Extends the recursive delete function, whenever a component was
	 * found within a span/anchor it will be moved in front of it. Empty
	 * spans/anchor will be deleted!
	 *
	 * ToDo: On demand cloning? Man traversiert den Baum durch und fuegt alles
	 * in ein temporaes Element ein? Oder gleich an den Anfang des Paragraphen?
	 * splitParagraph muesste die gleiche Funktionalitaet bekommen. Anchor wird
	 * nicht uebernommen!
	 */
//	private static void deleteMetadata(TextParagraphElementBase destinationComponent, int startPos, int endPos, boolean hasDeleteHyperlink, boolean hasDeleteStyle) {
//		System.out.println("deleting");
//	}
	private static void setParagraphListProperties(TextParagraphElementBase paragraphBaseElement, JSONObject paraProps, OdfFileDom xmlDoc, int listLevel, int newListLevel, String currentListStyleName) {
		// Will the listlevel change)
		if (listLevel != newListLevel) {
			// if there has been already a list
			if (listLevel > -1) {
				if (newListLevel > listLevel) {
					// raise the level of lists
					addListLevel(paragraphBaseElement, newListLevel - listLevel, paraProps, xmlDoc, currentListStyleName);
				} else {
					// lower the level of lists					
					removeListLevel(paragraphBaseElement, listLevel - newListLevel, paraProps, currentListStyleName);
				}
			} else { // there is no list yet (level plus one as it start counting with 0)
				addListLevel(paragraphBaseElement, newListLevel + 1, paraProps, xmlDoc, currentListStyleName);
			}
		}
		// Add list style and xml id to the top most list element				
		addTopListElementAttributes(getListRootElement(paragraphBaseElement), paraProps, null, currentListStyleName);
	}

	private static int getListLevel(TextParagraphElementBase paragraphBaseElement) {
		return getListLevel(paragraphBaseElement, -1);
	}

	/**
	 * Count the ancestor <text:list> elements to receive the list level
	 */
	private static int getListLevel(OdfElement element, int listLevel) {
		OdfElement parent = (OdfElement) element.getParentNode();
		if (parent instanceof TextListElement) {
			listLevel = getListLevel(parent, listLevel) + 1;
		} else if (parent instanceof TextListItemElement || parent instanceof TextListHeaderElement) {
			listLevel = getListLevel(parent, listLevel);
		}
		return listLevel;
	}

	/**
	 * Removal of list level starts bottom-up from the lowest list level. The
	 * reason is that by component position only the selected paragraph can be
	 * found. The paragraph element is nested within all list & list-items and
	 * has the deepest element position.
	 *
	 * If there the paragraphs has preceding and following list content the list
	 * has to be split. For this reason it the list will be cloned.
	 *
	 * As xml:id attributes have to be kept and are not being cloned, the clone
	 * will become the latter part and the original list the prior part.
	 *
	 * From the cloned part all content from the beginning including the marked
	 * paragraph are deleted.
	 *
	 * From the original part all content after the paragraph is being deleted.
	 *
	 * In addition if the paragraph is not the last, there has to be another
	 * clone. In the first clone all following paragraphs are being deleted, in
	 * the other last all preceding paragraphs are removed. The same procedure
	 * for paragraphs has to be undertaken for text:list-item & text:list-head
	 * elements!
	 */
	private static void removeListLevel(TextParagraphElementBase paragraphBaseElement, int levelsToDelete, JSONObject paraProps, String currentListStyleName) {
		// ToDo: Differentiate between list-header and list-item!!
		if (levelsToDelete > 0) {
			// Checks for preceding paragraphs, lists or list-items before the current list-item
			boolean hasPrecedingListContent = OdfElement.getPreviousElementSibling(paragraphBaseElement) != null || OdfElement.getPreviousElementSibling(paragraphBaseElement.getParentNode()) != null;
			// Checks for following paragraphs, lists or list-items after the current list-item			
			boolean hasFollowingListContent = OdfElement.getNextElementSibling(paragraphBaseElement) != null || OdfElement.getNextElementSibling(paragraphBaseElement.getParentNode()) != null;

			// if the paragraph is not the first and nor the last component, get the root of the list to clone the element
			// The parent of the paragraph is either a <text:list-item> or <text:list-header>, its parent is the desired <text:list> element			
			paragraphBaseElement.setAttribute("selectedParagraph", "true");
			OdfElement lowestListElement = (OdfElement) paragraphBaseElement.getParentNode().getParentNode();

			TextListElement clonedListWithFollowingContent = null;
			// if there has been content before the paragraph, clone this content
			if (hasPrecedingListContent) {
				if (hasFollowingListContent) {
					// *** CASE 3 ***: Dealing with following list content
					// clone current list for following content
					clonedListWithFollowingContent = (TextListElement) lowestListElement.cloneNode(true);
					Component selectedParagraphComponent = paragraphBaseElement.getComponent();
					int positionOfSelectedParagraph = selectedParagraphComponent.mParent.indexOf(selectedParagraphComponent);
					adjustComponentReferences(-1, positionOfSelectedParagraph, selectedParagraphComponent, clonedListWithFollowingContent);

					// only the FOLLOWING content is of interest, therefore delete everything BEFORE the selected paragraph										
					deleteBeforeAndSelectedParagraph(clonedListWithFollowingContent);
				}
				// *** CASE 1 & 3 ***: handling preceding list content
				// only the PRECEDING content is of interest, therefore delete everything AFTER the selected paragraph (including the paragraph)
				deleteFromSelectedParagraph(lowestListElement);

				boolean listLabelHidden = false;
				// if there is no previous sibling and hidden, we need a list-header
				if (paraProps.has("listLabelHidden")) {
					listLabelHidden = paraProps.optBoolean("listLabelHidden", Boolean.FALSE);
				}
				if (listLabelHidden) { // ToDo: We need automated tests!!
					OdfElement elementForInsertion = null;
					OdfElement newParentElement = null;

					// if there is no preceding content, we need to move the paragraph into a list-header element
					if (hasPrecedingListContent) {
						// if the list item is shown the pragraph has to be in a list-item
						newParentElement = (OdfElement) lowestListElement.getParentNode().getParentNode();
						// Create new List Item
						TextListHeaderElement newListHeaderElement = new TextListHeaderElement((OdfFileDom) paragraphBaseElement.getOwnerDocument());
						newListHeaderElement.appendChild(paragraphBaseElement);
						elementForInsertion = newListHeaderElement;

					} else {
						// if the list item is hidden it will be placed stand alone (or in a list-header at list beginning)
						newParentElement = (OdfElement) lowestListElement.getParentNode();
						// Create new List Item			
						elementForInsertion = paragraphBaseElement;
					}
					OdfElement nextListElementSibling = OdfElement.getNextElementSibling(lowestListElement);
					// if we are at the end of the list
					if (nextListElementSibling == null) {
						newParentElement.appendChild(elementForInsertion);
						if (clonedListWithFollowingContent != null) {
							newParentElement.appendChild(clonedListWithFollowingContent);
						}
					} else { // if we are in the middle of the list				
						newParentElement.insertBefore(elementForInsertion, nextListElementSibling);
						if (clonedListWithFollowingContent != null) {
							newParentElement.insertBefore(clonedListWithFollowingContent, nextListElementSibling);
						}
					}
				} else {
					boolean isTopLevelList = false;
					OdfElement elementForInsertion = null;
					OdfElement newParentElement = null;

					// parent is a list element 
					// as if the list item is shown the pragraph has to be in a list-item
					OdfElement listItemPrecedingSibling = (OdfElement) lowestListElement.getParentNode();
					OdfElement listItemFollowingSibling = null;
					if (listItemPrecedingSibling instanceof TextListItemElement || listItemPrecedingSibling instanceof TextListHeaderElement) {
						newParentElement = (OdfElement) listItemPrecedingSibling.getParentNode();
						listItemFollowingSibling = OdfElement.getNextElementSibling(listItemPrecedingSibling);
						// Create new List Item
						TextListItemElement newListItemElement = new TextListItemElement((OdfFileDom) paragraphBaseElement.getOwnerDocument());
						newListItemElement.appendChild(paragraphBaseElement);
						elementForInsertion = newListItemElement;
					} else {
						isTopLevelList = true;
						elementForInsertion = paragraphBaseElement;
						newParentElement = listItemPrecedingSibling;
						listItemFollowingSibling = OdfElement.getNextElementSibling(lowestListElement);
					}

//					OdfElement nextListElementSibling = OdfElement.getNextElementSibling(lowestListElement);
					// if we are at the end of the list
					if (listItemFollowingSibling == null) {
						newParentElement.appendChild(elementForInsertion);
						if (clonedListWithFollowingContent != null) {
							newParentElement.appendChild(clonedListWithFollowingContent);
						}

						if (clonedListWithFollowingContent != null) {
							if (isTopLevelList) {
								// Create new List Item
								newParentElement.appendChild(clonedListWithFollowingContent);

							} else {
								// Create new List Item
								TextListItemElement newListItemElement = new TextListItemElement((OdfFileDom) paragraphBaseElement.getOwnerDocument());
								newListItemElement.appendChild(clonedListWithFollowingContent);
								newParentElement.appendChild(newListItemElement);

							}
						}
					} else { // if we are in the middle of the list				
						newParentElement.insertBefore(elementForInsertion, listItemFollowingSibling);
						if (clonedListWithFollowingContent != null) {
							if (isTopLevelList) {
								// Create new List Item
								newParentElement.insertBefore(clonedListWithFollowingContent, listItemFollowingSibling);
							} else {
								// Create new List Item
								TextListItemElement newListItemElement = new TextListItemElement((OdfFileDom) paragraphBaseElement.getOwnerDocument());
								newListItemElement.appendChild(clonedListWithFollowingContent);
								newParentElement.insertBefore(newListItemElement, listItemFollowingSibling);
							}
						}
					}
				}
			} else {
				if (hasFollowingListContent) {
					// *** CASE 2 ***: Ony following list content
					// the list is moved after the paragraph, remove the paragraph out of the list
					deleteBeforeAndSelectedParagraph(lowestListElement);
					Node listParent = lowestListElement.getParentNode();
					listParent.insertBefore(paragraphBaseElement, lowestListElement);
				} else {
					// *** CASE 0 ***: ALONE - the paragraph was alone in the list parent, just replace the paragraph with list
					Node listParent = lowestListElement.getParentNode();
					listParent.replaceChild(paragraphBaseElement, lowestListElement);
					lowestListElement = null;
				}

			}



			//**** FIXING STYLES
			// if list element is a top list element (this happens if the list element has not list header/item as parent)
			if (lowestListElement != null && !(lowestListElement.getParentNode() instanceof TextListItemElement) && !(lowestListElement.getParentNode() instanceof TextListHeaderElement)) {
				OdfElement previousSiblingElement = OdfElement.getPreviousElementSibling(lowestListElement);
				String previousListStyleId = null;
				if (previousSiblingElement instanceof TextListElement) {
					previousListStyleId = ((TextListElement) previousSiblingElement).getTextStyleNameAttribute();
				}
				if (currentListStyleName == null) {
					currentListStyleName = ((TextListElement) lowestListElement).getTextStyleNameAttribute();
				}

				// Add list style and xml id to the top most list element	
				addTopListElementAttributes(getListRootElement(paragraphBaseElement), paraProps, previousListStyleId, currentListStyleName);

				if (clonedListWithFollowingContent != null) {
					// parameter properties have to be zero, as the existing styles should take precedence
					addTopListElementAttributes(clonedListWithFollowingContent, null, previousListStyleId, currentListStyleName);
				}
			}
			// removing another pairs of <text:list> and <text:list-item> or <text:list-header>			
			removeListLevel(paragraphBaseElement, --levelsToDelete, paraProps, currentListStyleName);
		} else {
			if (paragraphBaseElement.hasAttribute("selectedParagraph")) {
				paragraphBaseElement.removeAttribute("selectedParagraph");
			}
		}
	}

	private static void deleteBeforeAndSelectedParagraph(Element listElement) {
		listContentRemoval(listElement, true);
	}

	private static void deleteFromSelectedParagraph(Element listElement) {
		listContentRemoval(listElement, false);
	}

	/**
	 * Loops through elements of a list. There can only be list-item/header
	 * children and within those paragraphs. Within one paragraph an attribute
	 * was set. According to the given parameters the content before and/or
	 * after this paragraph is being deleted
	 */
	private static void listContentRemoval(Element listElement, boolean deleteBeforeSelection) {
		// deleteAfterSelection: All content after the selected paragraph in document order is being deleted
		// deleteBeforeSelection: The opposite, only one of both modes exists, the selected paragraph is always deleted
		boolean deleteAfterSelection = !deleteBeforeSelection;
		// searching backwards the list for the correct list item (to not corrupt the node list by removing its content while iterating)
		boolean isAfterSelectedParagraph = false;
		boolean isSelectedParagraph = false;
		NodeList listChildren = ((TextListElement) listElement).getChildNodes();
		for (int i = listChildren.getLength() - 1; i >= 0; i--) {
			Node listItem = listChildren.item(i);
			// check if it is really a list item/header and not for instance a text line break
			if (listItem instanceof Element) {
				NodeList listItemParagraphs = listItem.getChildNodes();
				// as the remove items from the node list we need to count back.. 
				for (int j = listItemParagraphs.getLength() - 1; j >= 0; j--) {
					Node listItemParagraph = listItemParagraphs.item(j);
					// check if it is really a list item/header and not for instance a text line break
					if (listItemParagraph instanceof OdfElement) {
						// If selected paragraph not found yet
						if (!isAfterSelectedParagraph) {
							// check if the content element is the selected paragraph
							if (((OdfElement) listItemParagraph).hasAttribute("selectedParagraph")) {
								isAfterSelectedParagraph = true;
								isSelectedParagraph = true;
								// the selected character will always be removed
								listItem.removeChild(listItemParagraph);
								// if no deletion before the seleciton is required..
								if (!deleteBeforeSelection) {
									// stop here..
									break;
								}
							} else if (deleteAfterSelection) {
								// IF DESIRED.. delete earlier content
								listItem.removeChild(listItemParagraph);
							}
						} else if (deleteBeforeSelection) {
							// IF DESIRED.. delete LATER content
							listItem.removeChild(listItemParagraph);
						}
					} else if (listItemParagraph instanceof Text) {
						// removing indentation
						listItem.removeChild(listItemParagraph);
					}
				}

				if (isAfterSelectedParagraph && !deleteBeforeSelection) {
//					// if the last content was removed..
//					if (listItemParagraphs.getLength() == 0) {
//						// remove the list itself
//						Element parent = (Element) listItem.getParentNode();
//						parent.removeChild(listItem);
//					}
					if (!listItem.hasChildNodes()) {
						// remove the list itself
						Element parent = (Element) listItem.getParentNode();
						parent.removeChild(listItem);
					}
					break;
				} else {
					// only do anything if it is not the selected paragraph
					if (!isSelectedParagraph) {
						// ..if we are still after
						if (!isAfterSelectedParagraph) {
							// ..and if the deletion mode is correct..
							if (deleteAfterSelection) {
								// ..delete list item
								listElement.removeChild(listItem);
							} else {
								if (!listItem.hasChildNodes()) {
									// remove the list itself
									Element parent = (Element) listItem.getParentNode();
									parent.removeChild(listItem);
								}
							}
							// else we are before and if the mode is correct..
						} else if (deleteBeforeSelection) {
							// delete list item
							listElement.removeChild(listItem);
						}
					} else {
						isSelectedParagraph = false;
						if (!listItem.hasChildNodes()) {
							// remove the list itself
							Element parent = (Element) listItem.getParentNode();
							parent.removeChild(listItem);
						}
					}
				}
			} else if (listItem instanceof Text) {
				// removing indentation
				listElement.removeChild(listItem);
			}
		}
	}

	/**
	 * All paragraph references from the original list to components that occur
	 * paragraphs AFTER the selected paragraph have to be moved to the
	 * clonedList paragraphs.
	 *
	 * @param originalListElement Within this list, every paragraph holds a
	 * reference to its component and vice versa
	 *
	 * @param clonedListElement Within this list, NO paragraph holds a reference
	 * to its component and vice versa *
	 */
	private static int adjustComponentReferences(int currentParagraphPosition, int selectedParagraphPosition, Component selectedParagraphComponent, Element clonedListElement) {
		NodeList listChildren = ((TextListElement) clonedListElement).getChildNodes();
		for (int i = 0; i < listChildren.getLength(); i++) {
			Node listItem = listChildren.item(i);
			// check if it is really a list item/header and not for instance a text line break
			if (listItem instanceof Element) {
				NodeList listItemParagraphs = listItem.getChildNodes();
				// as the remove items from the node list we need to count back.. 
				for (int j = 0; j < listItemParagraphs.getLength(); j++) {
					Node listItemParagraph = listItemParagraphs.item(j);
					// check if it is really a list item/header and not for instance a text line break
					if (listItemParagraph instanceof OdfElement) {
						// If selected paragraph not found yet
						if (listItemParagraph instanceof TextParagraphElementBase) {
							if (currentParagraphPosition == -1) {
								if (((OdfElement) listItemParagraph).hasAttribute("selectedParagraph")) {
									currentParagraphPosition = selectedParagraphPosition + 1;
								}
							} else if (currentParagraphPosition > selectedParagraphPosition) {
								// Add components to the paragraphs of the second part of the cloned list
								Component nextComponent = selectedParagraphComponent.mParent.getChild(currentParagraphPosition);
								nextComponent.setRootElement((TextParagraphElementBase) listItemParagraph);
								// adding back reference from element to component
								((TextParagraphElementBase) listItemParagraph).setComponent(nextComponent);
								currentParagraphPosition++;
							}
						} else if (listItemParagraph instanceof TextListElement) {
							currentParagraphPosition = adjustComponentReferences(currentParagraphPosition, selectedParagraphPosition, selectedParagraphComponent, (TextListElement) listItemParagraph);
						}
					}
				}
			}
		}
		return currentParagraphPosition;
	}

	/**
	 * In theory the list style might be set on any text:list within the nested
	 * list-items/headers, but all known offices (AO/LO/MSO) are consistently
	 * using the first style
	 */
	private static TextListElement getListRootElement(TextParagraphElementBase paragraphBaseElement) {
		OdfElement root = null;
		OdfElement parent = (OdfElement) paragraphBaseElement.getParentNode();
		while (parent instanceof TextListElement || parent instanceof TextListItemElement || parent instanceof TextListHeaderElement) {
			root = parent;
			parent = (OdfElement) parent.getParentNode();
		}
		return (TextListElement) root;
	}

	/**
	 * If a list already exists, new list level will be added inside of the
	 * existing list. In the first step of the recursion only the paragraphBase
	 * will be replaced by a list, further list will be added in the other
	 * addListLevel method recursive.
	 */
	private static void addListLevel(TextParagraphElementBase paragraphBaseElement, int levelsToCreate, JSONObject paraProps, OdfFileDom xmlDoc, String currentListStyleName) {
		// ToDo: Differentiate between list-header and list-item!!
		if (levelsToCreate > 0) {
			OdfElement parent = (OdfElement) paragraphBaseElement.getParentNode();
			// Create new List 
			TextListElement newListElement = new TextListElement(xmlDoc);

			// if list element is a top list element (this happens if the list element has not list header/item as parent)
			if (!(parent instanceof TextListItemElement) && !(parent instanceof TextListHeaderElement)) {
				OdfElement previousSiblingElement = OdfElement.getPreviousElementSibling(paragraphBaseElement);
				String previousListStyleId = null;
				if (previousSiblingElement instanceof TextListElement) {
					previousListStyleId = ((TextListElement) previousSiblingElement).getTextStyleNameAttribute();
				}
				// Add list style and xml id to the top most list element
				addTopListElementAttributes(newListElement, paraProps, previousListStyleId, currentListStyleName);
			}
			parent.replaceChild(newListElement, paragraphBaseElement);

			levelsToCreate--;
			if (levelsToCreate > 0) {
				// Create new List Item
				TextListItemElement newListItemElement = new TextListItemElement(xmlDoc);
				newListElement.appendChild(newListItemElement);
				newListItemElement.appendChild(paragraphBaseElement);
				addListLevel(paragraphBaseElement, levelsToCreate, paraProps, xmlDoc, currentListStyleName);
			} else {
				boolean listLabelHidden = false;;
				if (paraProps.has("listLabelHidden")) {
					listLabelHidden = paraProps.optBoolean("listLabelHidden", Boolean.FALSE);
				}
				if (listLabelHidden) {
					// Create new List Header			
					TextListHeaderElement newListHeaderElement = new TextListHeaderElement(xmlDoc);
					newListElement.appendChild(newListHeaderElement);
					newListHeaderElement.appendChild(paragraphBaseElement);
				} else {
					// Create new List Item
					TextListItemElement newListItemElement = new TextListItemElement(xmlDoc);
					newListElement.appendChild(newListItemElement);
					newListItemElement.appendChild(paragraphBaseElement);
				}
			}
		}
	}

	private static void addListLevel(OdfElement newParent, TextParagraphElementBase paragraphBaseElement, int levelsToCreate, JSONObject paraProps, OdfFileDom xmlDoc) {
		// ToDo: Differentiate between list-header and list-item!!
		if (levelsToCreate > 0) {
			// Create new List 
			TextListElement newListElement = new TextListElement(xmlDoc);
			newParent.appendChild(newListElement);

			if (levelsToCreate == 1) {
				boolean listLabelHidden = false;;
				if (paraProps.has("listLabelHidden")) {
					listLabelHidden = paraProps.optBoolean("listLabelHidden", Boolean.FALSE);
				}
				if (listLabelHidden) {
					// Create new List Item			
					TextListHeaderElement newListHeaderElement = new TextListHeaderElement(xmlDoc);
					newListElement.appendChild(newListHeaderElement);
					newListHeaderElement.appendChild(paragraphBaseElement);
				} else {
					// Create new List Item
					TextListItemElement newListItemElement = new TextListItemElement(xmlDoc);
					newListElement.appendChild(newListItemElement);
					newListItemElement.appendChild(paragraphBaseElement);
				}
			} else {
				// Create new List Item
				TextListItemElement newListItemElement = new TextListItemElement(xmlDoc);
				newListElement.appendChild(newListItemElement);

				levelsToCreate--;
				addListLevel(newListItemElement, paragraphBaseElement, levelsToCreate, paraProps, xmlDoc);
			}
		} else {
			newParent.appendChild(paragraphBaseElement);
		}
	}

	/**
	 * List styles should currently only changed for the paragraph being
	 * selected.
	 */
	private static void addTopListElementAttributes(TextListElement rootListElement, JSONObject paraProps, String previousSiblingListId, String currentListId) {
		if (paraProps != null && paraProps.length() != 0) {

			// Make certain the paragraph is in an list of its own, as only 						
			if (rootListElement != null) {
				String newListXmlId = null;
				String newListStyleId = null;
				if (paraProps.has("listXmlId")) {
					newListXmlId = paraProps.optString("listXmlId", null);
					if (newListXmlId != null) {
						if (!newListXmlId.isEmpty()) {
							// do not create an empty ID
							rootListElement.setXmlIdAttribute(newListXmlId);
						}
					} else {
						rootListElement.removeAttributeNS(OdfDocumentNamespace.XML.getUri(), "id");
					}
				}
				// If style or styleID provided are different to the existent
				if (paraProps.has("listStyleId")) {
					newListStyleId = paraProps.optString("listStyleId", null);
					if (newListStyleId != null) {
						if (!newListStyleId.isEmpty()) {
							rootListElement.setTextStyleNameAttribute(newListStyleId);
						} else {
							// reuse the previous list style ID
							if (currentListId == null && previousSiblingListId != null && !previousSiblingListId.isEmpty()) {
								rootListElement.setTextStyleNameAttribute(newListStyleId);
							}
						}
					} else {
						rootListElement.removeAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "style-name");
					}
				} else {
					if (currentListId == null && previousSiblingListId != null && !previousSiblingListId.isEmpty()) {
						rootListElement.setTextStyleNameAttribute(previousSiblingListId);
					} else if (currentListId != null && !currentListId.isEmpty()) {
						rootListElement.setTextStyleNameAttribute(currentListId);
					}
				}
				// Similar to MSO 15 ODF output & behavior, set by default the continue numbering to true
				rootListElement.setTextContinueNumberingAttribute(Boolean.TRUE);
			}
		} else {
			if (currentListId == null && previousSiblingListId != null && !previousSiblingListId.isEmpty()) {
				rootListElement.setTextStyleNameAttribute(previousSiblingListId);
			} else if (currentListId != null && !currentListId.isEmpty()) {
				rootListElement.setTextStyleNameAttribute(currentListId);
			}
			rootListElement.setTextContinueNumberingAttribute(Boolean.TRUE);
		}
	}

	public static void insertListStyle(String listStyleId, JSONObject listDefinition, OdfDocument doc) {
		try {
			// ToDo: Later in case a list was added to a header and footer a styles DOM have to be accessed
			OdfOfficeAutomaticStyles autoStyles = doc.getContentDom().getOrCreateAutomaticStyles();
			OdfTextListStyle listStyle = autoStyles.newListStyle(listStyleId);
			JSONObject listLevelDefinition = null;
			for (int i = 0; i < 9; i++) {
				listLevelDefinition = listDefinition.optJSONObject("listLevel" + i);
				addListDefinition(listStyle, listLevelDefinition, i);
			}
		} catch (Exception ex) {
			Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	/**
	 *
	 * <
	 * text:list-level-style-bullet text:level="1"
	 * text:style-name="Bullet_20_Symbols" text:bullet-char="\u2022">
	 * <style:list-level-properties
	 * text:list-level-position-and-space-mode="label-alignment">
	 * <style:list-level-label-alignment text:label-followed-by="listtab"
	 * text:list-tab-stop-position="1.27cm" fo:text-indent="-0.635cm"
	 * fo:margin-left="1.27cm"/>
	 * </style:list-level-properties>
	 * </text:list-level-style-bullet>
	 *
	 */
	private static void addListDefinition(OdfTextListStyle listStyle, JSONObject listLevelDefinition, int listLevel) {
		if (listLevelDefinition != null) {
			TextListLevelStyleElementBase listLevelStyle = null;

			// numberFormat: One of 'none', 'bullet', 'decimal', 'lowerRoman', 'upperRoman', 'lowerLetter', or 'upperLetter'.
			String numberFormat = listLevelDefinition.optString("numberFormat");
			String levelText = listLevelDefinition.optString("levelText");
			String numPrefix = null;
			String numSuffix = null;
			if (numberFormat.equals("bullet")) {
				// if there is also a suffix appended
				if (levelText.length() > 1) {
					// num-suffix
					numSuffix = levelText.substring(1);
					// bullet-char			
					levelText = levelText.substring(0, 1);
					// ToDo: API FIX to split prefix & suffix from bullet char, a single levelText will not be able to round-trip
				}
				String levelPicBulletUri = listLevelDefinition.optString("levelPicBulletUri");
				if (levelText != null) { // if there is no text label, we have to create an image list										
					if (levelPicBulletUri != null && !levelPicBulletUri.isEmpty()) {
						listLevelStyle = createListLevelStyleImage(listStyle, listLevel, levelPicBulletUri);
					} else {
						listLevelStyle = listStyle.newTextListLevelStyleBulletElement(levelText, listLevel + 1);
					}
				} else {
					listLevelStyle = createListLevelStyleImage(listStyle, listLevel, levelPicBulletUri);
				}
			} else { // *** NUMBERED LIST ***
				listLevelStyle = listStyle.newTextListLevelStyleNumberElement(getNumFormat(numberFormat), listLevel + 1);
				listLevelStyle.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:display-levels", Integer.toString(countOccurrences(levelText, '%')));

				// if there is prefix
				if (!levelText.startsWith("%")) {
					// num-prefix
					int prefixEnd = levelText.indexOf('%');
					numPrefix = levelText.substring(0, prefixEnd);
					levelText = levelText.substring(prefixEnd);
				}
				// num-suffix
				int suffixStart = levelText.lastIndexOf('%') + 2;
				if (levelText.length() >= suffixStart) {
					numSuffix = levelText.substring(suffixStart);
				}
				int listStartValue = listLevelDefinition.optInt("listStartValue", -1);
				if (listStartValue != -1) {
					listLevelStyle.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:start-value", Integer.toString(listStartValue));
				}
			}
			if (numPrefix != null && !numPrefix.isEmpty()) {
				listLevelStyle.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:num-prefix", numPrefix);
			}
			if (numSuffix != null && !numSuffix.isEmpty()) {
				listLevelStyle.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:num-suffix", numSuffix);
			}
			// Add common list style properties
			addCommonListStyles(listLevelStyle, listLevelDefinition);

		}
	}

	private static TextListLevelStyleElementBase createListLevelStyleImage(OdfTextListStyle listStyle, int listLevel, String levelPicBulletUri) {
		TextListLevelStyleElementBase listLevelStyle = listStyle.newTextListLevelStyleImageElement(listLevel + 1);

		if (levelPicBulletUri != null && !levelPicBulletUri.isEmpty()) {
			listLevelStyle.setAttributeNS(OdfDocumentNamespace.XLINK.getUri(), "xlink:href", levelPicBulletUri);
			TextListLevelStyleImageElement listLevelStyleImage = (TextListLevelStyleImageElement) listLevelStyle;
			listLevelStyleImage.setXlinkActuateAttribute("onLoad");
			listLevelStyleImage.setXlinkShowAttribute("embed");
			listLevelStyleImage.setXlinkTypeAttribute("simple");
		}
		return listLevelStyle;
	}

	private static int countOccurrences(String haystack, char needle) {
		int count = 0;
		for (int i = 0; i < haystack.length(); i++) {
			if (haystack.charAt(i) == needle) {
				count++;
			}
		}
		return count;
	}

	/**
	 * The <style:list-level-properties> element has the following attributes:
	 * <ul>
	 * <li>fo:height</li>
	 * <li>fo:text-align</li>
	 * <li>fo:width</li>
	 * <li>style:font-name</li>
	 * <li>style:vertical-pos</li>
	 * <li>style:vertical-rel</li>
	 * <li>svg:y</li>
	 * <li>text:list-level-position-and-space-mode</li>
	 * <li>text:min-label-distance</li>
	 * <li>text:min-label-width</li>
	 * <li>text:space-before</li>
	 * </ul>
	 */
	private static void addCommonListStyles(TextListLevelStyleElementBase listLevelStyle, JSONObject listLevelDefinition) {
		StyleListLevelPropertiesElement styleListLevelProperties = listLevelStyle.newStyleListLevelPropertiesElement();
		addListLabelAlignment(styleListLevelProperties, listLevelDefinition);

		if (listLevelDefinition.has("height")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:height", listLevelDefinition.optInt("height") / 100 + "mm");
		}
		if (listLevelDefinition.has("textAlign")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:text-align", listLevelDefinition.optString("textAlign"));
		}
		if (listLevelDefinition.has("width")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:width", listLevelDefinition.optInt("width") / 100 + "mm");
		}
		if (listLevelDefinition.has("font-name")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-name", listLevelDefinition.optString("fontName"));
		}
		if (listLevelDefinition.has("vertical-pos")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:vertical-pos", listLevelDefinition.optString("verticalPos"));
		}
		if (listLevelDefinition.has("vertical-rel")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:vertical-rel", listLevelDefinition.optString("verticalRel"));
		}
		if (listLevelDefinition.has("y")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.SVG.getUri(), "svg:y", listLevelDefinition.optString("y"));
		}
		if (listLevelDefinition.has("listLevelPositionAndSpaceMode")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:list-level-position-and-space-mode", listLevelDefinition.optString("listLevelPositionAndSpaceMode"));
		}
		if (listLevelDefinition.has("minLabelDistance")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:min-label-distance", listLevelDefinition.optInt("minLabelDistance") / 100 + "mm");
		}
		if (listLevelDefinition.has("minLabelWidth")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:min-label-width", listLevelDefinition.optInt("minLabelWidth") / 100 + "mm");
		}
		if (listLevelDefinition.has("spaceBefore")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:space-before", listLevelDefinition.optInt("spaceBefore") / 100 + "mm");
		}
	}

	private static void addListLabelAlignment(StyleListLevelPropertiesElement styleListLevelProperties, JSONObject listLevelDefinition) {
		if (listLevelDefinition.has("indentLeft") || listLevelDefinition.has("indentFirstLine")) {
			styleListLevelProperties.setAttributeNS(OdfDocumentNamespace.TEXT.getUri(), "text:list-level-position-and-space-mode", "label-alignment");
			String labelFollowedBy = listLevelDefinition.optString("labelFollowedBy", "listtab");
			StyleListLevelLabelAlignmentElement listLevelLabelAlignmentElement = styleListLevelProperties.newStyleListLevelLabelAlignmentElement(labelFollowedBy);

			if (listLevelDefinition.has("indentLeft")) {
				int indentLeft = listLevelDefinition.optInt("indentLeft");
				listLevelLabelAlignmentElement.setFoMarginLeftAttribute(indentLeft / 100.0 + "mm");
			}
			if (listLevelDefinition.has("indentFirstLine")) {
				int indentFirstLine = listLevelDefinition.optInt("indentFirstLine");
				listLevelLabelAlignmentElement.setFoTextIndentAttribute(indentFirstLine / 100.0 + "mm");
			}
			if (listLevelDefinition.has("tabStopPosition")) {
				int tabStopPosition = listLevelDefinition.optInt("tabStopPosition");
				listLevelLabelAlignmentElement.setTextListTabStopPositionAttribute(tabStopPosition / 100.0 + "mm");
				listLevelLabelAlignmentElement.setTextLabelFollowedByAttribute("listtab");
			}
		}
	}

	/**
	 * The style:num-format attribute specifies a numbering sequence. The
	 * defined values for the style:num-format attribute are: 1: Hindu-Arabic
	 * number sequence starts with 1. a: number sequence of lowercase Modern
	 * Latin basic alphabet characters starts with "a". A: number sequence of
	 * uppercase Modern Latin basic alphabet characters starts with "A". i:
	 * number sequence of lowercase Roman numerals starts with "i". I: number
	 * sequence of uppercase Roman numerals start with "I". a value of type
	 * string. an empty string: no number sequence displayed. If no value is
	 * given, no number sequence is displayed.
	 */
	private static String getNumFormat(String numberFormat) {

		String numFormat = ""; // "none" is nothing set
		if (numberFormat.equals("decimal")) {
			numFormat = "1";
		} else if (numberFormat.equals("lowerRoman")) {
			numFormat = "i";
		} else if (numberFormat.equals("upperRoman")) {
			numFormat = "I";
		} else if (numberFormat.equals("lowerLetter")) {
			numFormat = "a";
		} else if (numberFormat.equals("upperLetter")) {
			numFormat = "A";
		}
		return numFormat;
	}

	/**
	 * Splits a paragraph into two splitting the text among the two. Special
	 * handling for paragraph within lists: In this case the list item will be
	 * split!
	 */
	public static void splitParagraph(Component rootComponent, JSONArray start) throws IndexOutOfBoundsException {
		Component component = rootComponent.getParentOf(start);
		if (component != null) {
			TextParagraphElementBase para = (TextParagraphElementBase) component.getRootElement();
			try {
				TextParagraphElementBase newSecondPara = (TextParagraphElementBase) para.split(start.getInt(start.length() - 1));
				Component parentComponent = component.getParent();
				Node paraParent = component.getRootElement().getParentNode();

				// if paragraph within a list (checked on parent of list-item or list-header) the split will not only split the paragraph, but as well this item/header
				// BECAUSE: In contrary to ODF in OOXML the list is only a property of a paragraph, there can only be one paragraph within a list (emulating this behavior).
				if (paraParent instanceof TextListItemElement || paraParent instanceof TextListHeaderElement) {
					OdfElement listItem = (OdfElement) paraParent;
					// split the list inbetween the old (splitted) and new paragraph, plus one as a new paragraph was added above 
					int precedingSiblingsNo = para.countPrecedingElementSiblings() + 1;
					OdfElement secondListPart = listItem.split(precedingSiblingsNo);
					parentComponent.createChildComponent(start.getInt(start.length() - 2) + 1, secondListPart.getFirstElementChild());
				} else {
					parentComponent.createChildComponent(start.getInt(start.length() - 2) + 1, newSecondPara);
				}
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		} else {
			LOG.severe("Could not find paragraph at position: " + start);
		}
	}

//    public static void mergeParagraph(Component rootComponent, int paraPos) throws IndexOutOfBoundsException {
//        OdfTextParagraph newParent = (OdfTextParagraph) rootComponent.getChildRoot(paraPos);
//        if (newParent != null) {
//            OdfTextParagraph oldParent = (OdfTextParagraph) rootComponent.getChildRoot(paraPos + 1);
//            if (oldParent != null) {
//
//                oldParent.moveChildrenTo(newParent);
//            } else {
//                throw new IndexOutOfBoundsException("The index " + paraPos + 1 + " was accessed as source of the merge. The last index is " + (rootComponent.size() - 1));
//            }
//        } else {
//            throw new IndexOutOfBoundsException("The index " + paraPos + " was accessed as destination of the merge. The last index is " + (rootComponent.size() - 1));
//        }
//    }
	private static void deleteText(Component rootComponent, JSONArray start, JSONArray end) throws IndexOutOfBoundsException {
		Component parentComponent = rootComponent.getParentOf(start);
		if (parentComponent != null) {
			//TextParagraphElementBase startParagraph = (TextParagraphElementBase) parentComponent.getRootElement();
			try {
				// ToDo - add the paragraph text length to the paragraph? Than I need to intercept all text manipulations!! *ARGH just do-it! :D
				// ToDo - parse the text spans positions from the beginning, creating array - could do this from the beginning..
				// better not for very long texts it is unnecessary, does not scale
				//        adding the span depths, the span element referece, for every character?			
				if (parentComponent instanceof TextComponent) {
					((TextComponent) parentComponent).removeText(start.getInt(start.length() - 1), end.getInt(end.length() - 1));
				} else {
					LOG.log(Level.SEVERE, "The parent of the text is not a text component: {0}", parentComponent);
				}
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		} else {
			LOG.log(Level.SEVERE, "Could not find paragraph as parent for position: {0}", start);
		}
	}

	public static int applyOperations(TextOperationDocument opsDoc, JSONArray ops) throws Exception {
		LOG.log(Level.FINEST, "*************** INCOMING OPERATIONS **********\n{0}", ops);
		int acceptedOpCount = 0;
		OdfDocument doc = opsDoc.getDocument();
		Component rootComponent = doc.getRootComponent();
		JSONObject operation = null;
		try {
			for (int i = 0; i < ops.length(); i++) {
				operation = (JSONObject) ops.get(i);
				String opName = operation.getString("name");
				if (opName.equals("insertParagraph")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONObject attrs = operation.optJSONObject("attrs");
					insertParagraph(rootComponent, start, attrs);
				} else if (opName.equals("delete")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONArray end = operation.optJSONArray("end");
					delete(rootComponent, start, end);
				} else if (opName.equals("insertText")) {// missing the mapping of whitespaces
					acceptedOpCount++;
					JSONObject attrs = operation.optJSONObject("attrs");
					JSONArray start = operation.getJSONArray("start");
					String text = operation.getString("text");
					insertText(rootComponent, start, attrs, text);
				} else if (opName.equals("deleteText")) { // missing the mapping of whitespaces
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONArray end = operation.optJSONArray("end");
					delete(rootComponent, start, end);
				} else if (opName.equals("setAttributes")) {
					acceptedOpCount++;
					JSONObject attrs = operation.optJSONObject("attrs");
					JSONArray start = operation.getJSONArray("start");
					JSONArray end = operation.optJSONArray("end");
					setAttributes(rootComponent, start, end, attrs);
				} else if (opName.equals("mergeParagraph")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					mergeParagraph(rootComponent, start);
				} else if (opName.equals("splitParagraph")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					splitParagraph(rootComponent, start);
				} else if (opName.equals("insertTable")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONObject attrs = operation.optJSONObject("attrs");
					JSONObject sizeExceeded = operation.optJSONObject("sizeExceeded");
					insertTable(rootComponent, start, attrs, sizeExceeded);
				} else if (opName.equals("insertRows")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONObject attrs = operation.optJSONObject("attrs");
					int count = operation.optInt("count", 1);
					boolean insertDefaultCells = operation.optBoolean("insertDefaultCells");
					int referenceRow = operation.optInt("referenceRow", -1);
					insertRows(rootComponent, start, attrs, count, insertDefaultCells, referenceRow);
				} else if (opName.equals("insertCells")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONObject attrs = operation.optJSONObject("attrs");
					int count = operation.optInt("count", 1);
					insertCells(rootComponent, start, attrs, count, null, null);
				} else if (opName.equals("insertStyleSheet")) {
					acceptedOpCount++;
					String type = operation.getString("type");
					String styleId = operation.getString("styleId");
					String styleName = operation.optString("styleName");
					JSONObject attrs = operation.optJSONObject("attrs");
					String parent = operation.optString("parent");
					Boolean hidden = operation.optBoolean("hidden");
					insertStyleSheet(doc, type, styleId, styleName, attrs, parent, hidden);
				} else if (opName.equals("deleteStyleSheet")) {
					acceptedOpCount++;
					String type = operation.getString("type");
					String styleId = operation.getString("styleId");
					deleteStyleSheet(doc, styleId, type);
				} else if (opName.equals("insertColumn")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					Integer gridPosition = operation.getInt("gridPosition");
					JSONArray tableGrid = operation.getJSONArray("tableGrid");
					String insertMode = operation.optString("insertMode");
					insertColumn(rootComponent, start, gridPosition, tableGrid, insertMode);
				} else if (opName.equals("deleteColumns")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					Integer startGrid = operation.getInt("startGrid");
					Integer endGrid = operation.optInt("endGrid");
					deleteColumns(rootComponent, start, startGrid, endGrid);
				} else if (opName.equals("insertFontDescription")) {
					acceptedOpCount++;
					JSONObject attrs = operation.getJSONObject("attrs");
					JSONArray panose1 = attrs.optJSONArray("panose1");
					String panose1Value = null;
					if (panose1 != null) {
						panose1Value = panose1.toString();
					}
					String fontName = operation.getString("fontName");
					String[] altNames = (String[]) attrs.opt("altNames");
					String family = attrs.optString("family");
					String familyGeneric = attrs.optString("familyGeneric");
					String pitch = attrs.optString("pitch");
					insertFontDescription(doc, fontName, altNames, family, familyGeneric, pitch, panose1Value);
				} else if (opName.equals("insertTab")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONObject attrs = operation.optJSONObject("attrs");
					insertTab(rootComponent, start, attrs);
				} else if (opName.equals("insertHardBreak")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONObject attrs = operation.optJSONObject("attrs");
					insertHardBreak(rootComponent, start, attrs);
				} else if (opName.equals("insertDrawing")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					String type = operation.optString("type");
					JSONObject attrs = operation.optJSONObject("attrs");
					insertDrawing(rootComponent, start, attrs, type, opsDoc);
				} else if (opName.equals("insertField")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					String representation = operation.optString("representation");
					insertField(rootComponent, start, representation);
				} else if (opName.equals("move")) {
					acceptedOpCount++;
					JSONArray start = operation.getJSONArray("start");
					JSONArray to = operation.getJSONArray("to");
					//	JSONArray end = operation.optJSONArray("end");
					moveDrawing(rootComponent, start, to);
				} else if (opName.equals("insertListStyle")) {
					acceptedOpCount++;
					String listStyle = operation.optString("listStyleId");
					JSONObject listDefinition = operation.optJSONObject("listDefinition");
					insertListStyle(listStyle, listDefinition, doc);
				}
			}
		} catch (Exception e) {
			LOG.log(Level.SEVERE, "An error occured in the operation number {0}{1}", new Object[]{acceptedOpCount, 1});
			LOG.severe("The operation was: " + operation);
			throw e;
		}
		return acceptedOpCount;
	}

	private static void insertFontDescription(OdfDocument doc, String fontName, String[] altNames, String family, String familyGeneric, String pitch, String panose1) {
		// CREATING CONTENT FONT DECLARATION
		try {
			OdfContentDom contentDom = doc.getContentDom();
			StyleFontFaceElement newElement = new StyleFontFaceElement(contentDom);
			newElement.setStyleNameAttribute(fontName);
			if (family != null && !family.isEmpty()) {
				newElement.setSvgFontFamilyAttribute(family);
			}
			if (familyGeneric != null && !familyGeneric.isEmpty()) {
				newElement.setStyleFontFamilyGenericAttribute(familyGeneric);
			}
			if (pitch != null && !pitch.isEmpty()) {
				newElement.setStyleFontPitchAttribute(pitch);
			}
			if (panose1 != null && !panose1.isEmpty()) {
				if (panose1.contains("[")) {
					panose1 = panose1.substring(1, panose1.length() - 1);
				}
				if (panose1.contains(",")) {
					panose1 = panose1.replace(',', ' ');
				}
				newElement.setSvgPanose1Attribute(panose1);
			}
			insertFontFace(contentDom, newElement);

			OdfStylesDom stylesDom = doc.getStylesDom();
			StyleFontFaceElement newElement2 = (StyleFontFaceElement) newElement.cloneNode(true);
			newElement2 = (StyleFontFaceElement) stylesDom.adoptNode(newElement2);
			insertFontFace(stylesDom, newElement2);

		} catch (Exception ex) {
			Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	private static void insertFontFace(OdfFileDom fileDom, StyleFontFaceElement fontFaceElement) {
		NodeList fontDecls = fileDom.getElementsByTagName("office:font-face-decls");
		if (fontDecls.getLength() == 1) {
			OfficeFontFaceDeclsElement fontDecl = (OfficeFontFaceDeclsElement) fontDecls.item(0);
			fontDecl.appendChild(fontFaceElement);
		} else {
			Element rootElement = fileDom.getRootElement();
			NodeList rootChildren = rootElement.getChildNodes();
			boolean hasInserted = false;
			for (int i = 0; i < rootChildren.getLength(); i++) {
				Node currentNode = rootChildren.item(i);
				if (currentNode instanceof Element) {
					String elementName = ((Element) currentNode).getNodeName();
					if (elementName.equals("office:automatic-styles") || elementName.equals("office:styles") || elementName.equals("office:body") || elementName.equals("office:master-styles")) {
						OfficeFontFaceDeclsElement fontFaceDeclsElement = new OfficeFontFaceDeclsElement(fileDom);
						fontFaceDeclsElement.appendChild(fontFaceElement);
						rootElement.insertBefore(fontFaceDeclsElement, currentNode);
						hasInserted = true;
						break;
					}
				} else {
					continue;
				}
			}
			if (!hasInserted) {
				rootElement.appendChild(fontFaceElement);
			}
		}
	}

	/**
	 * Currently all fields are being mapped to a common text field, i.e.
	 * <text:text-input>
	 */
	public static void insertField(Component rootComponent, JSONArray start, String representation) {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);
		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			// CREATING NEW ROOT ELEMENT		
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			TextTextInputElement newFieldElement = new TextTextInputElement(xmlDoc);
			if (representation != null) {
				Text text = xmlDoc.createTextNode(representation);
				newFieldElement.appendChild(text);
			}
			try {
				// ADDING COMPONENT
				addComponents(parentComponent, newFieldElement, start.getInt(start.length() - 1), 1);
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
	}

	public static void insertDrawing(Component rootComponent, JSONArray start, JSONObject attrs, String type, TextOperationDocument opsDoc) throws IndexOutOfBoundsException {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);
		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			boolean isImageFrame = (type != null && type.equals("image"));
			// CREATING NEW ROOT ELEMENT		
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			DrawFrameElement frameElement = null;
			JSONObject drawingProps = null;
			int alternativeNo = -1;
			if (attrs != null) {
				drawingProps = (JSONObject) attrs.optJSONObject("drawing");
				if (drawingProps != null) {
					if (drawingProps.has("viewAlternative")) {
						try {
							alternativeNo = drawingProps.optInt("viewAlternative");
							frameElement = (DrawFrameElement) parentComponent.getChild(start.getInt(start.length() - 1)).getRootElement();
						} catch (JSONException ex) {
							Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
						}
					} else {
						frameElement = new DrawFrameElement(xmlDoc);
					}
					if (frameElement != null) {
						setFrameProperties(frameElement, drawingProps);
					}
				}
				// ADDING STYLES TO THE NEW ELEMENT
				addStyle(attrs, frameElement, xmlDoc);
			}
			DrawImageElement imageElement = null;
			if (isImageFrame) {
				imageElement = frameElement.newDrawImageElement();
				imageElement.setXlinkTypeAttribute("simple");
				imageElement.setXlinkShowAttribute("embed");
				imageElement.setXlinkActuateAttribute("onLoad");
				if (drawingProps != null) {
					setImageProperties(imageElement, drawingProps);
				}
			}

			// INSERTING THE IMAGE
			try {
				if (imageElement != null) {
					if (alternativeNo == -1) {
						// ADDING COMPONENT				
						addComponents(parentComponent, frameElement, start.getInt(start.length() - 1), 1);
					} else {
						NodeList frameChildren = frameElement.getChildNodes();
						int imagesFound = 0;
						Node followingImage = null;
						for (int i = 0; i < frameChildren.getLength(); i++) {
							Node node = frameChildren.item(i);
							if (node instanceof DrawImageElement) {
								imagesFound++;
							}
							if (imagesFound > alternativeNo) {
								followingImage = node;
								break;
							}
						}
						// if followingImage is null, the imageElement will be appended (see API)
						frameElement.insertBefore(imageElement, followingImage);
					}
				}
			} catch (JSONException ex) {
				Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
			}

			// INSERTING THE IMAGE TO THE ZIP
			if (drawingProps.has("imageUrl") && imageElement != null && imageElement instanceof OdfDrawImage) {
				String packagePath;
				try {
					packagePath = drawingProps.getString("imageUrl");
					if (packagePath.contains("uid")) {
						int uidStart = packagePath.indexOf("uid") + 3;
						if (uidStart != 3 && uidStart < packagePath.length() && uidStart < packagePath.indexOf('.')) {
							String uidString = null;
							if (packagePath.contains(".")) {
								uidString = packagePath.substring(uidStart, packagePath.indexOf('.'));
							} else {
								uidString = packagePath.substring(uidStart);
							}
							if (uidString != null) {
								long uid = Long.parseLong(uidString, 16);
								Map<Long, byte[]> resourceMap = opsDoc.getResourceMap();
								if (resourceMap != null) {
									byte[] fileBytes = resourceMap.get(uid);
									if (fileBytes != null) {
										OdfDrawImage image = (OdfDrawImage) imageElement;
										try {
											image.newImage(fileBytes, packagePath, OdfFileEntry.getMediaTypeString(packagePath));
										} catch (Exception ex) {
											Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
										}
									}
								}
							}
						} else {
							LOG.severe("No appropriate resource URL found for picture: " + packagePath);
						}
					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
				}
			}
		}
	}

	private static void setImageProperties(DrawImageElement imageElement, JSONObject drawingProps) {
		if (drawingProps.has("imageUrl")) {
			String href = drawingProps.optString("imageUrl");
			if (href != null && !href.isEmpty()) {
				imageElement.setXlinkHrefAttribute(href);
			}
		} else if (drawingProps.has("imageData")) {
			String imageData = drawingProps.optString("imageData");
			if (imageData != null && !imageData.isEmpty()) {
				//expected header is
				//		"data:image/png;base64,
				String[] header = imageData.split("base64,");
				String suffix = "png";
				String mediaTypeString = "image/png";
				try {
					mediaTypeString = header[0].substring(header[0].indexOf(":") + 1, header[0].indexOf(";"));
					suffix = header[0].substring(header[0].indexOf("/") + 1, header[0].indexOf(";"));
				} catch (Throwable t) {
					// don't worry if the header is not as expected
					LOG.finer("BASE64 is not as expected '" + header[0] + "', the exception:" + t);
				}
				String fileName = "img_" + new Random().nextInt() + "." + suffix;
				Base64Binary base64Binary = Base64Binary.valueOf(header[1]);
				OdfPackage pkg = ((OdfFileDom) imageElement.getOwnerDocument()).getDocument().getPackage();
				// args: fileBytes, internalPath, mediaTypeString) {
				pkg.insert(base64Binary.getBytes(), "Pictures/" + fileName, mediaTypeString);
				imageElement.setXlinkHrefAttribute("Pictures/" + fileName);
			}
		}
		if (drawingProps.has("imageXmlId")) {
			String xmlId = drawingProps.optString("imageXmlId");
			if (xmlId != null && !xmlId.isEmpty()) {
				imageElement.setXmlIdAttribute(xmlId);
			}
		}
	}

	private static void setFrameProperties(DrawFrameElement frameElement, JSONObject drawingProps) {
		if (drawingProps.has("width")) {
			int width = drawingProps.optInt("width");
			frameElement.setSvgWidthAttribute(width / 100.0 + "mm");
		}
		if (drawingProps.has("height")) {
			int height = drawingProps.optInt("height");
			frameElement.setSvgHeightAttribute(height / 100.0 + "mm");
		}
		if (drawingProps.has("name")) {
			String name = drawingProps.optString("name");
			if (name != null && !name.isEmpty()) {
				frameElement.setDrawNameAttribute(name);
			}
		}
		if (drawingProps.has("anchorHorOffset")) {
			int x = drawingProps.optInt("anchorHorOffset");
			frameElement.setSvgXAttribute(x / 100.0 + "mm");
		}
		if (drawingProps.has("anchorVertOffset")) {
			int y = drawingProps.optInt("anchorVertOffset");
			frameElement.setSvgYAttribute(y / 100.0 + "mm");
		}
		if (drawingProps.has("anchorHorBase") && drawingProps.has("anchorVertBase")) {
			String anchorHorBase = drawingProps.optString("anchorHorBase");
			String anchorVertBase = drawingProps.optString("anchorVertBase");
			if (anchorHorBase != null && anchorVertBase != null) {
				if (anchorHorBase.equals("page") && anchorVertBase.equals("page")) {
					frameElement.setTextAnchorTypeAttribute("page");

				} else if (anchorHorBase.equals("column") && anchorVertBase.equals("margin")) {
					frameElement.setTextAnchorTypeAttribute("frame");

				} else if (anchorHorBase.equals("column") && anchorVertBase.equals("paragraph")) {
					frameElement.setTextAnchorTypeAttribute("paragraph");

				} else if (anchorHorBase.equals("character") && anchorVertBase.equals("paragraph")) {
					frameElement.setTextAnchorTypeAttribute("char");
				}
			}
		} else {
			if (drawingProps.has("anchorHorBase")) {
				String anchorHorBase = drawingProps.optString("anchorHorBase");
				if (anchorHorBase != null) {
					if (anchorHorBase.equals("page")) {
						frameElement.setTextAnchorTypeAttribute("page");

					} else if (anchorHorBase.equals("column")) {
						frameElement.setTextAnchorTypeAttribute("paragraph");

					} else if (anchorHorBase.equals("character")) {
						frameElement.setTextAnchorTypeAttribute("char");
					}
				}
			} else {
				frameElement.setTextAnchorTypeAttribute("as-char");
			}

		}
	}

	/**
	 * @param rootComponent high level document structure
	 * @param pos position of the table starting with 0 and one over the
	 * existing is allowed
	 * @throws IndexOutOfBoundsException - if index is out of range (index < 0
	 * || index > size()). One over size is allowed to append a table.
	 */
	public static void insertTable(Component rootComponent, JSONArray start, JSONObject attrs, JSONObject sizeExceeded) throws IndexOutOfBoundsException {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);
		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			OdfStylableElement newTableElement;
			if (sizeExceeded == null) {
				// CREATING NEW ROOT ELEMENT		
				newTableElement = new TableTableElement(xmlDoc);

				// ADDING STYLES TO THE NEW ELEMENT
				addStyle(attrs, newTableElement, xmlDoc);
				try {
					if (attrs != null) {
						JSONObject tableProps = attrs.getJSONObject("table");
						JSONArray tableGrid = tableProps.getJSONArray("tableGrid");
						insertColumns((TableTableElement) newTableElement, tableGrid);
					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
				}

			} else {
				newTableElement = new TextPElement(xmlDoc);
				newTableElement.appendChild(xmlDoc.createTextNode("This table was too large!"));				
			}
				try {
					// ADDING COMPONENT
					addComponents(parentComponent, newTableElement, start.getInt(start.length() - 1), 1);
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
				}			
		}
	}

	private static void setAbsoluteColumnWidths(TableTableElement tableElement, JSONArray tableGrid) {
		List<TableTableColumnElement> columnList = getTableColumnElements(tableElement, new LinkedList<TableTableColumnElement>());

		List<String> absColumnWidths = null;
		if (tableGrid != null) {
			absColumnWidths = getAbsoluteColumnWidths(tableElement, tableGrid);
		}

		// EXCHANGE ALL ABSOLUTE WIDTHS WITH THE GIVEN RELATIVE WIDTHS
		// We are iterating the elements from the columnList, but receive the width from the grid
		for (int elementNo = 0, columnNo = 0; elementNo < columnList.size(); elementNo++, columnNo++) {
			TableTableColumnElement columnElement = columnList.get(elementNo);
			StyleStyleElement columnStyleElement = columnElement.getOrCreateUnqiueAutomaticStyle();
			StyleTableColumnPropertiesElement columnPropsElement = (StyleTableColumnPropertiesElement) columnStyleElement.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableColumnProperties);
			int repeated = -1;
			if (columnPropsElement == null) {
				columnPropsElement = columnStyleElement.newStyleTableColumnPropertiesElement();
			} else if (columnElement.hasAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-columns-repeated")) {
				String repeatedValue = columnElement.getAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-columns-repeated");
				if (repeatedValue != null && !repeatedValue.isEmpty()) {
					repeated = Integer.parseInt(repeatedValue);
					long relColumnWidth = tableGrid.optLong(columnNo);
					if (absColumnWidths != null) {
						columnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:column-width", absColumnWidths.get(columnNo));
					} else {
						// if there is no absolute table width, the relative widths have to be used (or "auto" being resolved)
						columnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:rel-column-width", tableGrid.optLong(columnNo) + "*");
					}
					int repeatedColumn = 0;
					TableTableColumnElement lastColumnElement = columnElement;
					OdfElement nextColumnSibling = null;
					for (int i = 1; i <= repeated; i++, columnNo++) {
						long nextColumnWidth = tableGrid.optLong(columnNo);
						if (relColumnWidth != nextColumnWidth) {
							// clone the original column element
							TableTableColumnElement newColumnElement = (TableTableColumnElement) lastColumnElement.cloneNode(true);
							// adapt existing repeated attribute							
							lastColumnElement.setTableNumberColumnsRepeatedAttribute(repeatedColumn);
							newColumnElement.setTableNumberColumnsRepeatedAttribute(1);
							// Retrieve the follow up node of the repeated column only once
							if (nextColumnSibling == null) {
								nextColumnSibling = OdfElement.getNextElementSibling(columnElement);
							}
							tableElement.insertBefore(newColumnElement, nextColumnSibling);


							// add the new width
							StyleStyleElement newColumnStyleElement = newColumnElement.getOrCreateUnqiueAutomaticStyle();
							StyleTableColumnPropertiesElement newColumnPropsElement = (StyleTableColumnPropertiesElement) newColumnStyleElement.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableColumnProperties);
							if (absColumnWidths != null) {
								newColumnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:column-width", absColumnWidths.get(columnNo));
							} else {
								// if there is no absolute table width, the relative widths have to be used (or "auto" being resolved)
								newColumnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:rel-column-width", tableGrid.optLong(columnNo) + "*");
							}
							// remember the latest values as the last
							relColumnWidth = nextColumnWidth;
							lastColumnElement = newColumnElement;
							repeatedColumn = 1;
						} else {
							repeatedColumn++;
						}
					}
					lastColumnElement.setTableNumberColumnsRepeatedAttribute(repeatedColumn);
				}
			} else if (absColumnWidths != null) {
				columnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:column-width", absColumnWidths.get(columnNo));
			} else {
				// if there is no absolute table width, the relative widths have to be used (or "auto" being resolved)
				columnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:rel-column-width", tableGrid.optLong(columnNo) + "*");
			}
		}
	}

	private static List<String> getAbsoluteColumnWidths(TableTableElement tableElement, JSONArray tableGrid) {
		List<String> absColumnWidths = null;
		String tableWidth = tableElement.getProperty(StyleTablePropertiesElement.Width);
		if (tableWidth != null && !tableWidth.isEmpty()) {
			int absTableWidth = normalizeLength(tableWidth);
			double relTableWidth = 0.0;
			int columnCount = tableGrid.length();
			absColumnWidths = new ArrayList(columnCount);
			for (int i = 0; columnCount > i; i++) {
				relTableWidth += tableGrid.optLong(i);
			}
			if (relTableWidth == 0) {
				// should never occur, but being defensive here, before dividing through zero 
				relTableWidth = 100;
			}
			double widthProRel = absTableWidth / relTableWidth;
			for (int i = 0; tableGrid.length() > i; i++) {
				double absWidth = Math.round(tableGrid.optInt(i) * widthProRel);
				absColumnWidths.add(absWidth / 100.0 + "mm");
			}
		}
		return absColumnWidths;
	}

	/**
	 * Inserts cells for a new column into a table.
	 *
	 * @param rootComponent high level document structure
	 * @param start	Integer[]	The logical position of the table the new column
	 * will be inserted into.
	 * @param gridPosition	Integer	Zero-based column index, according to the
	 * tableGrid attribute of the table.
	 * @param tableGrid	Integer[]	The complete array of relative widths for the
	 * entire table, containing the new entry for the new column. Will be set to
	 * the tableGrid attribute of the table.
	 * @param insertMode	String	(optional) If set to 'before', the new cells
	 * will be inserted before the existing cells, otherwise after the cells.
	 * Default is 'before'. Note: As a side effect, this operation changes the
	 * tableGrid attribute of the parent table.
	 */
	public static void insertColumn(Component rootComponent, JSONArray start, Integer gridCellSourcePosition, JSONArray tableGrid, String insertMode) {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.get(start);
		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The table parent component of the column should exist at position {0}", start);
		} else {
			// MULTIPLYING THE COLUMN
			TableTableElement tableElement = (TableTableElement) parentComponent.getRootElement();
			List<TableTableColumnElement> columnList = getTableColumnElements(tableElement, new LinkedList<TableTableColumnElement>());
			boolean isRepeatedReferenceColumn = false;
			List<String> absColumnWidths = null;
			if (tableGrid != null) {
				absColumnWidths = getAbsoluteColumnWidths(tableElement, tableGrid);
			}

			// EXCHANGE ALL ABSOLUTE WIDTHS WITH THE GIVEN RELATIVE WIDTHS
			// We are iterating the elements from the columnList, but receive the width from the grid
			for (int elementNo = 0, columnNo = 0; elementNo < columnList.size(); elementNo++, columnNo++) {
				TableTableColumnElement columnElement = columnList.get(elementNo);
				StyleStyleElement columnStyleElement = columnElement.getOrCreateUnqiueAutomaticStyle();
				StyleTableColumnPropertiesElement columnPropsElement = (StyleTableColumnPropertiesElement) columnStyleElement.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableColumnProperties);
				int repeated = -1;
				if (columnPropsElement == null) {
					columnPropsElement = columnStyleElement.newStyleTableColumnPropertiesElement();
				}
				if (absColumnWidths != null) {
					columnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:column-width", absColumnWidths.get(elementNo));
				} else {
					// if there is no absolute table width, the relative widths have to be used (or "auto" being resolved)
					columnPropsElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:rel-column-width", tableGrid.optLong(elementNo) + "*");
				}
				// skip the iteration over the same column
				if (columnElement.hasAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-columns-repeated")) {
					String repeatedValue = columnElement.getAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-columns-repeated");
					if (repeatedValue != null && !repeatedValue.isEmpty()) {
						repeated = Integer.parseInt(repeatedValue);
						// if the current column is equal the requested column position OR
						// if the current column is less or equal the current position + the repeated minus one (as 1 means no repetition)
						if (columnNo == gridCellSourcePosition || gridCellSourcePosition <= columnNo + repeated - 1) {
							// Already adust repeated instead of later insertion
							columnElement.setAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "table:number-columns-repeated", String.valueOf(repeated + 1));
							isRepeatedReferenceColumn = true;
						}
						if (repeated > 1) {
							columnNo += repeated - 1;
						}
					}
				}
			}

			// INSERTING THE NEW COLUMN
			int cellDestination = -1;
			if (!isRepeatedReferenceColumn) {
				TableTableColumnElement origColumnElement = columnList.get(gridCellSourcePosition);
				TableTableColumnElement newColumnElement = (TableTableColumnElement) origColumnElement.cloneNode(true);
				// do not clone an ID, as there can only be one of it in an XML file
				if (newColumnElement.hasAttribute("xml:id")) {
					newColumnElement.removeAttribute("xml:id");
				}

				if (insertMode.equals("before")) {
					OdfElement parent = (OdfElement) origColumnElement.getParentNode();
					parent.insertBefore(newColumnElement, origColumnElement);
					cellDestination = gridCellSourcePosition;
				} else { // if the column should be inserted after the given
					OdfElement parent = (OdfElement) origColumnElement.getParentNode();
					if (columnList.size() > gridCellSourcePosition + 1) {
						OdfElement nextSibling = columnList.get(gridCellSourcePosition + 1);
						parent.insertBefore(newColumnElement, nextSibling);
						cellDestination = gridCellSourcePosition + 1;
					} else {
						parent.appendChild(newColumnElement);
					}
				}
			} else { // repeated attribute has been already previously enhanced plus one
				if (insertMode.equals("before")) {
					cellDestination = gridCellSourcePosition;
				} else {
					cellDestination = gridCellSourcePosition + 1;
				}
			}


//				StyleStyleElement origColumnStyleElement = origColumnElement.getOrCreateUnqiueAutomaticStyle();
//				StyleTableColumnPropertiesElement origColumnPropsElement = (StyleTableColumnPropertiesElement) columnStyleElement.getPropertiesElement(OdfStylePropertiesSet.TableColumnProperties);
//				StyleTableColumnPropertiesElement newColumnPropsElement;
//				if (origColumnPropsElement != null) {
//					newColumnPropsElement = (StyleTableColumnPropertiesElement) origColumnStyleElement.getPropertiesElement(OdfStylePropertiesSet.TableColumnProperties);
//					String origColumnWidth = origColumnPropsElement.getStyleColumnWidthAttribute();
//					if (origColumnWidth != null && !origColumnWidth.isEmpty()) {
//						double origWidthCM = new Length(origColumnWidth).getCentimeter();
//						long newWidth = Math.round(origWidthCM / origRelColumnWidth * 10000.0 * newRelColumnWidth); // origWidth / newRelWidth = orig
//						newColumnPropsElement.setStyleColumnWidthAttribute(String.valueOf(newWidth / 10000.0) + "cm");
//					}
//				} else {
//					newColumnPropsElement = newColumnElement.getOrCreateUnqiueAutomaticStyle().newStyleTableColumnPropertiesElement();
//					newColumnPropsElement = (StyleTableColumnPropertiesElement) columnStyleElement.newStyleTableColumnPropertiesElement();
//					// ToDo: Fallback of any other column width absolute style or the table itself
//				}
//				newColumnPropsElement.setStyleRelColumnWidthAttribute(newRelColumnWidth + "*");


			// MULTIPLYING THE CELLS - adding XML & Component
			int currentRowPosition = 0;
			int rowComponentLevel = start.length();
			// Add inital row and cell position
			JSONArray destinationPosition = start.put(currentRowPosition).put(cellDestination);
			for (Object o : parentComponent.getChildren()) {
				Component rowComponent = (Component) o;
				TableTableCellElement clonedCellElement = (TableTableCellElement) rowComponent.getChild(gridCellSourcePosition).getRootElement().cloneNode(false);
				Integer columnRepeated = clonedCellElement.getTableNumberColumnsRepeatedAttribute();
				if (columnRepeated != null && columnRepeated > 1) {
					clonedCellElement.removeAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-columns-repeated");
				}
				Integer columnSpanned = clonedCellElement.getTableNumberColumnsSpannedAttribute();
				if (columnSpanned != null && columnSpanned > 1) {
					clonedCellElement.removeAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-columns-spanned");
				}
				if (clonedCellElement.hasAttribute("xml:id")) {
					clonedCellElement.removeAttribute("xml:id");
				}
				insertCells(rootComponent, destinationPosition, null, 1, rowComponent, clonedCellElement);
				try {
					currentRowPosition++;
					destinationPosition.put(rowComponentLevel, currentRowPosition);
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
				}
			}
		}
	}

	private static List<TableTableColumnElement> getTableColumnElements(Element parent, List columns) {
		NodeList children = parent.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			if (child instanceof Element) {
				if (child instanceof TableTableColumnElement) {
					columns.add((TableTableColumnElement) child);
				} else if (child instanceof TableTableColumnGroupElement
						|| child instanceof TableTableHeaderColumnsElement
						|| child instanceof TableTableColumnsElement) {
					columns = getTableColumnElements((Element) child, columns);
				} else if (child instanceof TableTableRowGroupElement
						|| child instanceof TableTableHeaderRowsElement
						|| child instanceof TableTableRowElement
						|| child instanceof TableTableRowsElement) {
					break;
				}
			}
		}
		return columns;
	}

	private static List<TableTableRowElement> getTableRowElements(Element parent, List rows) {
		NodeList children = parent.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			if (child instanceof Element) {
				if (child instanceof TableTableRowElement) {
					rows.add((TableTableRowElement) child);
				} else if (child instanceof TableTableRowGroupElement
						|| child instanceof TableTableHeaderRowsElement) {
					rows = getTableRowElements((Element) child, rows);
				}
			}
		}
		return rows;
	}

	/**
	 * Deletes the cells of a single column or multiple columns from a table.
	 *
	 * @param rootComponent high level document structure
	 * @param start	Integer[]	The logical position of the table whose columns
	 * will be removed.
	 * @param startGrid	Integer	Zero-based index of the first column to be
	 * removed, according to the tableGrid attribute of the table.
	 * @param endGrid	Integer	(optional) Zero-based index of the last column to
	 * be removed (closed range). If omitted, only one column will be removed.
	 * Note: Each cell that is addressed by the specified column range will be
	 * removed completely. As a side effect, this operation changes the
	 * tableGrid attribute of the parent table.
	 */
	public static void deleteColumns(Component rootComponent, JSONArray start, Integer startGrid, Integer endGrid) {
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.get(start);
		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The table parent component of the column should exist at position {0}", start);
		} else {
			// CREATING NEW ROOT ELEMENT		
			TableTableElement tableElement = (TableTableElement) parentComponent.getRootElement();
			OdfTable table = OdfTable.getInstance(tableElement);
			table.removeColumnsByIndex(startGrid, endGrid - startGrid + 1);
		}
	}

	/**
	 * Adding a sequence of columns with repeated attribute to the table and
	 * automatic styles with relative column width for the columns
	 */
	private static void insertColumns(TableTableElement table, JSONArray tableGrid) {
		if (tableGrid != null) {
			int columnLength = Integer.MIN_VALUE;
			int previousColumnLength = Integer.MIN_VALUE;
			int repeated = 1;
			int columnCount = tableGrid.length();
			if (tableGrid != null) {
				for (int i = 0; i < columnCount; i++) {
					try {
						columnLength = (Integer) tableGrid.get(i);
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
					if (previousColumnLength == columnLength) {
						repeated++;
					} else {
						if (i != 0) {
							// write out column with repeat
							TableTableColumnElement column = table.newTableTableColumnElement();
							if (repeated != 1) {
								column.setTableNumberColumnsRepeatedAttribute(repeated);
							}
							column.getOrCreateUnqiueAutomaticStyle().newStyleTableColumnPropertiesElement().setStyleRelColumnWidthAttribute(previousColumnLength + "*");
						}
						previousColumnLength = columnLength;
						repeated = 1;
					}
				}
				TableTableColumnElement column = table.newTableTableColumnElement();
				if (repeated != 1) {
					column.setTableNumberColumnsRepeatedAttribute(repeated);
				}
				column.getOrCreateUnqiueAutomaticStyle().newStyleTableColumnPropertiesElement().setStyleRelColumnWidthAttribute(previousColumnLength + "*");
			}
		}
	}

	/**
	 * insertRows Inserts one or more new rows into a table.
	 *
	 * @param name	'insertRows'
	 * @param start	The logical position of the new row. The row will be
	 * inserted before a row that is currently located at this position.
	 * @param count	(optional) The number of rows that will be inserted, default
	 * is 1.
	 * @param insertDefaultCells	(optional) If true, empty cells will be
	 * inserted into the new row. The number of inserted cells will be equal to
	 * the number of columns in the table, as specified by the tableGrid
	 * attribute of the table. The default is false.
	 * @param referenceRow	(optional) If specified, the zero-based index of the
	 * existing row whose cells and their attributes will be cloned (but without
	 * the cell contents).
	 * @param attrs	(optional) Initial row attributes. See Table Row Formatting
	 * Attributes. Note: The attributes insertDefaultCells and referenceRow are
	 * mutually exclusive.
	 */
	public static void insertRows(Component rootComponent, JSONArray start, JSONObject attrs, int count, boolean insertDefaultCells, int referenceRow) throws IndexOutOfBoundsException, JSONException {
		assert count >= 1 : "Row count is expected to be at least 1";
		// Parent will not change and have to exist to insert the new component
		final Component parentComponent = rootComponent.getParentOf(start);
		TableTableElement tableElement = (TableTableElement) parentComponent.getRootElement();
		OdfTable table = OdfTable.getInstance(tableElement);

		if (parentComponent == null) {
			LOG.log(Level.SEVERE, "The parent component of the table should exist at position {0}", start);
		} else {
			TableTableRowElement rowOriginalElement;
			TableTableRowElement rowElement;
			OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
			if (referenceRow > -1) {
				// Get the referenceRow and copy it WITH cells, but without the cell content 
				OdfTableRow row = table.getRowByIndex(referenceRow);
				rowOriginalElement = row.getOdfElement();

				// Once cloned to apply all the hard styles only on the new row
				rowElement = (TableTableRowElement) rowOriginalElement.cloneNode(1);
			} else {
				// CREATING NEW ROOT ELEMENT		
				rowOriginalElement = new TableTableRowElement(xmlDoc);
				rowElement = new TableTableRowElement(xmlDoc);

			}
			// do not clone an ID, as there can only be one of it in an XML file
			if (rowElement.hasAttribute("xml:id")) {
				rowElement.removeAttribute("xml:id");
			}
			Integer rowsRepeated = rowElement.getTableNumberRowsRepeatedAttribute();
			if (rowsRepeated != null && rowsRepeated > 1) {
				rowElement.removeAttributeNS(OdfDocumentNamespace.TABLE.getUri(), "number-rows-repeated");
			}
			// ADDING STYLES TO THE NEW ELEMENT
			addStyle(attrs, rowElement, xmlDoc);

			// get the existing row to insert before
			Component referenceRowComponent = rootComponent.get(start);
			TableTableRowElement previousRow = null;
			if (referenceRowComponent != null) {
				// insert new row before
				if (referenceRowComponent.mRootElement != null) {
					previousRow = (TableTableRowElement) referenceRowComponent.mRootElement;
				}
			}

			// Insert row as often as requested by "count"
			// ToDO: Able to generalize XML and component adding?
			for (int i = 0; i < count; i++) { // ADD ROWS
				Component newRowComponent = addComponents(parentComponent, rowElement, -1, 1);

				// Only add cells to a new row, when exlicitly requested or reference row had been addressed
				if (referenceRow > -1) {
					NodeList rowChildren = rowElement.getChildNodes();
					Node child;

					for (int m = 0; m < rowChildren.getLength(); m++) { // ADD CELL ELEMENTS
						child = rowChildren.item(m);
						if (child instanceof TableTableCellElement) {
							// remove all the content from the cell (or covered table cell)
							TableTableCellElement cell = (TableTableCellElement) child;
							// ToDo: ADD EXISTING CELL TO THE COMPONENT TREE (PARSE EINEN ODF SUBTREE MIT OFFSET!!)
							// OTHERWISE WITH THE CLONING ROOT ELMENTS / COMPONENTS MIGHT BE ADDED!!!!
							// ToDo: ADD COUNTER TO COMPONENT TO AVOID EXPLIZIT REPEATED ROWS/CELLS

							for (int k = cell.getTableNumberColumnsRepeatedAttribute(); k > 0; k--) { // ADD CELL COMPONENTS
								newRowComponent.createChildComponent((OdfElement) cell);
							}
						}
					}
				} else if (insertDefaultCells) {
					// Fill the new row several times
					int cellCount = table.getColumnCount();
					insertCells(rootComponent, start, CELL_WITH_BORDER_ATTRS, cellCount, newRowComponent, null);
				}
				if (previousRow != null) {
					// insert new row before
					tableElement.insertBefore(rowElement, previousRow);
				} else {
					// append new row					
					tableElement.appendChild(rowElement);
				}

				// Only cloning again, if another row is required. First cloning happened for hard styling before the loop
				if (i < count) {
					rowElement = (TableTableRowElement) rowOriginalElement.cloneNode(1);
					// ToDo: There has to be a better way to avoid the cloning of rows with cells! Perhaps Component cloning?! :D
					addStyle(attrs, rowElement, xmlDoc);
				}
			}
		}
	}

// ToDo: insertCells/Rows/Table are very very SIMILAR and can be condensed!!
	public static void insertCells(Component rootComponent, JSONArray start, JSONObject attrs, int count, Component parentComponent, OdfElement cellElement) throws IndexOutOfBoundsException {
		// Parent will not change and have to exist to insert the new component
		if (parentComponent == null) {
			parentComponent = rootComponent.getParentOf(start);
		}
//
//		if(parentComponent == null){
//			// OOStyledTable.odt
//			// tableOps.odt
//		}
//		if (parentComponent == null) {
//			parentComponent = rootComponent.getParentOf(start);
//		}		
		OdfFileDom xmlDoc = (OdfFileDom) parentComponent.getOwnerDocument();
		if (cellElement == null) {
			// Creating a new cell (with border for text documents)
			cellElement = new TableTableCellElement(xmlDoc);
		}

		// Adding any hard formatting to cell
		if (attrs != null) {
			addStyle(attrs, (OdfStylableElement) cellElement, xmlDoc);
			addCellAttributes(attrs, (TableTableCellElement) cellElement);
		}

		// ADDING COMPONENT
		int cellPosition = start.optInt(start.length() - 1);
		addComponents(parentComponent, cellElement, cellPosition, count);
	}

	private static void addCellAttributes(JSONObject attrs, TableTableCellElement newCellElement) { //, OdfFileDom ownerDocument
		JSONObject props = (JSONObject) attrs.opt("cell");
		if (props != null && props.length() > 0) {
			if (props.has("gridSpan")) {
				Integer gridSpan = props.optInt("gridSpan");
				newCellElement.setTableNumberColumnsSpannedAttribute(gridSpan);
			}
		}
	}

	private static Component addComponents(Component parentComponent, OdfElement newElement, int newPosition, int count) {
		Component newComponent = null;
		Component existingComponent = null;

		// Mark the element as component, so for instance an ODF Frame with image can be recognized as component
		newElement.markAsComponentRoot(true);

		// CHECK IF AN EXISTING COMPONENT HAVE TO BE MOVED	
		// ToDo Later: It should work to proof first, if the rootElement length is less the newPosition..		
		if (newPosition != -1 && (existingComponent = parentComponent.getChild(newPosition)) != null) {
			// if there is already a component on this position, insert the new component before
			Element existingElement = existingComponent.getRootElement();
			Element parentElement = (Element) parentComponent.getRootElement();
			// PLACING THE NEW COMPONENT INTO THE EXISTING TREE		
			for (int i = 0; i < count; i++) {
				parentElement.insertBefore(newElement, existingElement);
				newComponent = parentComponent.createChildComponent(newPosition, newElement);
			}
		} else {
			if (parentComponent instanceof TextComponent && ((TextComponent) parentComponent).getComponentNode(newPosition) != null) {
				Element parentElement = (Element) parentComponent.getRootElement();
				if (parentElement instanceof OdfElement) {
					((OdfElement) parentElement).insert(newElement, newPosition);
				} else {
					LOG.warning("The parent element " + parentElement.getTagName() + " is not of type OdfElement. The new element " + newElement.getTagName() + " could not be inserted!");
				}
			} else {
				// ELSE THE COMPONENT IS THE NEW LAST AND HAVE TO BE APPENDED TO THE PARENT								
				// if there is not yet a component on this position, append the component to the parent
				Element parentElement = (Element) parentComponent.getRootElement();
				parentElement.appendChild(newElement);
				newComponent = parentComponent.createChildComponent(newElement);
				// if multiple components have to be added, clone them (different handling components with repeated element, e.g. for cells )
				for (int i = 1; i < count; i++) {
					newElement = (OdfElement) newElement.cloneNode(true);
					// do not clone an ID, as there can only be one of it in an XML file
					if (newElement.hasAttribute("xml:id")) {
						newElement.removeAttribute("xml:id");
					}
					parentElement.appendChild(newElement);
					newComponent = parentComponent.createChildComponent(newElement);
				}
			}
		}
		return newComponent;
	}

	static private void mapProperties(OdfStyleFamily styleFamily, JSONObject attrs, OdfStyleBase style, OdfDocument doc4Fonts) {

		if (attrs != null && styleFamily != null && style != null) {
			Map<String, OdfStylePropertiesSet> familyProperties = Component.getAllOxStyleGroupingIdProperties(styleFamily);
			Set<String> propTypes = familyProperties.keySet();
			for (String type : propTypes) {
				if (type.equals("character")) {
					JSONObject textProps = (JSONObject) attrs.opt("character");
					if (textProps != null && textProps.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.TextProperties);
						mapCharacterProperties(textProps, (StyleTextPropertiesElement) propsElement, doc4Fonts);
					}
				} else if (type.equals("paragraph")) {
					JSONObject paraProps = (JSONObject) attrs.opt("paragraph");
					if (paraProps != null && paraProps.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.ParagraphProperties);
						mapParagraphProperties(paraProps, (StyleParagraphPropertiesElement) propsElement);
					}
				} else if (type.equals("table")) {
					JSONObject tableProps = (JSONObject) attrs.opt("table");
					if (tableProps != null && tableProps.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableProperties);
						mapTableProperties(tableProps, (StyleTablePropertiesElement) propsElement);
					}
				} else if (type.equals("row")) {
					JSONObject props = (JSONObject) attrs.opt("row");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableRowProperties);
						mapRowProperties(props, (StyleTableRowPropertiesElement) propsElement);
					}
				} else if (type.equals("cell")) {
					JSONObject props = (JSONObject) attrs.opt("cell");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableCellProperties);
						mapCellProperties(props, (StyleTableCellPropertiesElement) propsElement);
					}
				} else if (type.equals("column")) {
					JSONObject props = (JSONObject) attrs.opt("column");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.TableColumnProperties);
						mapColumnProperties(props, (StyleTableColumnPropertiesElement) propsElement);
					}
				} else if (type.equals("list")) {
					JSONObject props = (JSONObject) attrs.opt("list");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.ListLevelProperties);
						mapListProperties(props, (StyleListLevelPropertiesElement) propsElement);
					}
				} else if (type.equals("section")) {
					JSONObject props = (JSONObject) attrs.opt("section");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.SectionProperties);
						mapSectionProperties(props, (StyleSectionPropertiesElement) propsElement);
					}

				} else if (type.equals("drawing") || type.equals("presentation")) {
					JSONObject props = (JSONObject) attrs.opt("drawing");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.GraphicProperties);
						mapGraphicProperties(props, (StyleGraphicPropertiesElement) propsElement);
					}
// ToDo: How to differentiate Graphics from Drawings? ( see condition before - images have only GraphicProperties!)					
//				} else if (type.equals("drawing")) {
//					JSONObject props = (JSONObject) attrs.opt("drawing");
//					if (props != null && props.length() > 0) {
//						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.DrawingPageProperties);
//						mapDrawingProperties(props, (StyleDrawingPagePropertiesElement) propsElement);
//					}					
				} else if (type.equals("chart")) {
					JSONObject props = (JSONObject) attrs.opt("chart");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.ChartProperties);
						mapChartProperties(props, (StyleChartPropertiesElement) propsElement);
					}
				} else if (type.equals("page")) {
					JSONObject props = (JSONObject) attrs.opt("page");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.PageLayoutProperties);
						mapPageProperties(props, (StylePageLayoutPropertiesElement) propsElement);
					}
				} else if (type.equals("ruby")) {
					JSONObject props = (JSONObject) attrs.opt("ruby");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.RubyProperties);
						mapRubyProperties(props, (StyleRubyPropertiesElement) propsElement);
					}
				} else if (type.equals("headerFooter")) {
					JSONObject props = (JSONObject) attrs.opt("headerFooter");
					if (props != null && props.length() > 0) {
						OdfStylePropertiesBase propsElement = style.getOrCreatePropertiesElement(OdfStylePropertiesSet.HeaderFooterProperties);
						mapHeaderFooterProperties(props, (StyleHeaderFooterPropertiesElement) propsElement);
					}
				}
			}
		}
	}

	// The latter fontNames Set is a hack for the OX release as the 16 fonts supported do not write out their descriptions
	static private void mapCharacterProperties(JSONObject attrs, StyleTextPropertiesElement propertiesElement, OdfDocument doc4Fonts) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				// TODO -- !!!!!!!!!!!!!!!!ROUNDTRIP WITH THESE VALUES!!!!!!!!!!!
				//	<define name="fontWeight">
				//		<choice>
				//			<value>normal</value>
				//			<value>bold</value>
				//			<value>100</value>
				//			<value>200</value>
				//			<value>300</value>
				//			<value>400</value>
				//			<value>500</value>
				//			<value>600</value>
				//			<value>700</value>
				//			<value>800</value>
				//			<value>900</value>
				//		</choice>
				//	</define>						
				// fo:font-weight
				if (key.equals("bold")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-weight");
						} else {
							Boolean isBold = (Boolean) value;
							if (isBold) {
								propertiesElement.setFoFontWeightAttribute("bold");
							} else {
								propertiesElement.setFoFontWeightAttribute("normal");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("boldAsian")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-weight-asian");
						} else {
							Boolean isBold = (Boolean) value;
							if (isBold) {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-weight-asian", "bold");
							} else {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-weight-asian", "normal");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("boldComplex")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-weight-complex");
						} else {
							Boolean isBold = (Boolean) value;
							if (isBold) {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-weight-complex", "bold");
							} else {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-weight-complex", "normal");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} //	<define name="lineStyle">
				//		<choice>
				//			<value>none</value>
				//			<value>solid</value>
				//			<value>dotted</value>
				//			<value>dash</value>
				//			<value>long-dash</value>
				//			<value>dot-dash</value>
				//			<value>dot-dot-dash</value>
				//			<value>wave</value>
				//		</choice>
				//	</define>				
				else if (key.equals("underline")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-underline-style");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-underline-width");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-underline-color");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-underline-type");
						} else {
							Boolean isUnderline = (Boolean) value;
							if (isUnderline) {
								propertiesElement.setStyleTextUnderlineStyleAttribute("solid");
							} else {
								propertiesElement.setStyleTextUnderlineStyleAttribute("none");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}

					//	<define name="fontStyle">
					//		<choice>
					//			<value>normal</value>
					//			<value>italic</value>
					//			<value>oblique</value>
					//		</choice>
					//	</define>					
				} else if (key.equals("italic")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-style");
						} else {
							Boolean isItalic = (Boolean) value;
							if (isItalic) {
								propertiesElement.setFoFontStyleAttribute("italic");
							} else {
								propertiesElement.setFoFontStyleAttribute("normal");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("italicAsian")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-style-asian");
						} else {
							Boolean isItalic = (Boolean) value;
							if (isItalic) {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-style-asian", "italic");

							} else {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-style-asian", "normal");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("italicComplex")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-style-complex");
						} else {
							Boolean isItalic = (Boolean) value;
							if (isItalic) {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-style-complex", "italic");
							} else {
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-style-complex", "normal");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("color")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "color");
						} else {
							JSONObject color = (JSONObject) value;
							if (color.hasAndNotNull("type")) {
								String type = color.getString("type");
								if (!type.equals(AUTO)) {
									propertiesElement.setFoColorAttribute(getColor(color, null));
									propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "use-window-font-color");
								} else {
									propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:use-window-font-color", "true");
								}
							} else { // DEFAULT IS AUTO
								propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:use-window-font-color", "true");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontSize")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "font-size");
						} else {
							Integer fontSize = (Integer) value;
							propertiesElement.setFoFontSizeAttribute(fontSize.toString() + "pt");
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontSizeAsian")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-size-asian");
						} else {
							Integer fontSize = (Integer) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-size-asian", fontSize.toString() + "pt");
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontSizeComplex")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-size-complex");
						} else {
							Integer fontSize = (Integer) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-size-complex", fontSize.toString() + "pt");
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontName")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-name");
						} else {
							String fontName = (String) value;
							propertiesElement.setStyleFontNameAttribute(fontName);
							addFontToDocument(fontName, doc4Fonts);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontNameAsian")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-name-asian");
						} else {
							String fontName = (String) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-name-asian", fontName);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontNameComplex")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-name-complex");
						} else {
							String fontName = (String) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-name-complex", fontName);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("vertAlign")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-position");
						} else {
							String alignment = (String) value;
							if (alignment.equals("sub")) {
								propertiesElement.setStyleTextPositionAttribute("sub");
							} else if (alignment.equals("super")) {
								propertiesElement.setStyleTextPositionAttribute("super");
							} else { //baseline
								propertiesElement.setStyleTextPositionAttribute("0% 100%");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("strike")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-position");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-color");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-mode");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-style");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-text");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-text-style");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-type");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "text-line-through-width");
						} else {
							String strikeType = (String) value;
							if (strikeType.equals("single")) {
								propertiesElement.setStyleTextLineThroughTypeAttribute("single");
								propertiesElement.setStyleTextLineThroughStyleAttribute("solid");
								propertiesElement.setStyleTextLineThroughModeAttribute("continuous");
								propertiesElement.setStyleTextUnderlineModeAttribute("continuous");
								propertiesElement.setStyleTextOverlineModeAttribute("continuous");
							} else if (!strikeType.equals("none")) { //double
								propertiesElement.setStyleTextLineThroughTypeAttribute("double");
								propertiesElement.setStyleTextLineThroughStyleAttribute("solid");
								propertiesElement.setStyleTextLineThroughModeAttribute("continuous");
								propertiesElement.setStyleTextUnderlineModeAttribute("continuous");
								propertiesElement.setStyleTextOverlineModeAttribute("continuous");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("language")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "country");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "language");
						} else {
							String locale = (String) value;
							if (!locale.isEmpty()) {
								int delimiterPos = locale.indexOf('-');
								if (delimiterPos > -1) {
									propertiesElement.setFoLanguageAttribute(locale.substring(0, delimiterPos));
									propertiesElement.setFoCountryAttribute(locale.substring(delimiterPos + 1, locale.length()));
								} else {
									propertiesElement.setFoLanguageAttribute(locale);
								}
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else /* 
				 The defined values for the fo:line-height attribute are:
				 * a value of type nonNegativeLength 
				 * normal: disables the effects of style:line-height-at-least and style:line-spacing.  
				 * a value of type percent  */ //				<attribute name="fo:line-height">
				//					<choice>
				//						<value>normal</value>
				//						<ref name="nonNegativeLength"/>
				//						<ref name="percent"/>
				//					</choice>
				//				</attribute>
				// { type: 'percent', value: 100 }																
				if (key.equals("letterSpacing")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "letter-spacing");
						} else {
							if (value.equals("normal")) {
								propertiesElement.setFoLetterSpacingAttribute("normal");
							} else {
								propertiesElement.setFoLetterSpacingAttribute(((Integer) value) / 100.0 + "mm");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}
//	DEFAULT FONTS ADDED TO APACHE OPEN OFFICE
//		<style:font-face style:name="Andale Mono" svg:font-family="&apos;Andale Mono&apos;"/>
//		<style:font-face style:name="Mangal1" svg:font-family="Mangal"/>
//		<style:font-face style:name="Palatino" svg:font-family="Palatino" style:font-family-generic="roman"/>
//		<style:font-face style:name="Times" svg:font-family="Times" style:font-family-generic="roman"/>
//		<style:font-face style:name="Helvetica" svg:font-family="Helvetica" style:font-family-generic="swiss"/>
//		<style:font-face style:name="Consolas" svg:font-family="Consolas" style:font-family-generic="modern" style:font-pitch="fixed"/>
//		<style:font-face style:name="Courier" svg:font-family="Courier" style:font-family-generic="modern" style:font-pitch="fixed"/>
//		<style:font-face style:name="Courier New" svg:font-family="&apos;Courier New&apos;" style:font-family-generic="modern" style:font-pitch="fixed"/>
//		<style:font-face style:name="Book Antiqua" svg:font-family="&apos;Book Antiqua&apos;" style:font-family-generic="roman" style:font-pitch="variable"/>
//		<style:font-face style:name="Cambria" svg:font-family="Cambria" style:font-family-generic="roman" style:font-pitch="variable"/>
//		<style:font-face style:name="Georgia" svg:font-family="Georgia" style:font-family-generic="roman" style:font-pitch="variable"/>
//		<style:font-face style:name="Times New Roman" svg:font-family="&apos;Times New Roman&apos;" style:font-family-generic="roman" style:font-pitch="variable"/>
//		<style:font-face style:name="Arial" svg:font-family="Arial" style:font-family-generic="swiss" style:font-pitch="variable"/>
//		<style:font-face style:name="Calibri" svg:font-family="Calibri" style:font-family-generic="swiss" style:font-pitch="variable"/>
//		<style:font-face style:name="Impact" svg:font-family="Impact" style:font-family-generic="swiss" style:font-pitch="variable"/>
//		<style:font-face style:name="Tahoma" svg:font-family="Tahoma" style:font-family-generic="swiss" style:font-pitch="variable"/>
//		<style:font-face style:name="Verdana" svg:font-family="Verdana" style:font-family-generic="swiss" style:font-pitch="variable"/>
//		<style:font-face style:name="Mangal" svg:font-family="Mangal" style:font-family-generic="system" style:font-pitch="variable"/>
//		<style:font-face style:name="Microsoft YaHei" svg:font-family="&apos;Microsoft YaHei&apos;" style:font-family-generic="system" style:font-pitch="variable"/>
//		<style:font-face style:name="SimSun" svg:font-family="SimSun" style:font-family-generic="system" style:font-pitch="variable"/>

// ADDED FROM MSO 15 export
/*
	 <style:font-face style:name="Cambria" svg:font-family="Cambria" style:font-family-generic="roman" style:font-pitch="variable" svg:panose-1="2 4 5 3 5 4 6 3 2 4"/>
	 <style:font-face style:name="MS Mincho" svg:font-family="MS Mincho" style:font-family-generic="modern" style:font-pitch="fixed" svg:panose-1="2 2 6 9 4 2 5 8 3 4"/>
	 <style:font-face style:name="Times New Roman" svg:font-family="Times New Roman" style:font-family-generic="roman" style:font-pitch="variable" svg:panose-1="2 2 6 3 5 4 5 2 3 4"/>
	 <style:font-face style:name="Calibri" svg:font-family="Calibri" style:font-family-generic="swiss" style:font-pitch="variable" svg:panose-1="2 15 5 2 2 2 4 3 2 4"/>
	 <style:font-face style:name="MS Gothic" svg:font-family="MS Gothic" style:font-family-generic="modern" style:font-pitch="fixed" svg:panose-1="2 11 6 9 7 2 5 8 2 4"/>
  
	 */
	private static void addFontToDocument(String fontName, OdfDocument doc4Fonts) {
		Set fontNames = doc4Fonts.getFontNames();
		if (fontName != null && !fontName.isEmpty()) {
			if (!fontNames.contains(fontName)) {
				fontNames.add(fontName);
				if (fontName.equals("Andale Mono")) {
					insertFontDescription(doc4Fonts, "Andale Mono", null, "Andale Mono", null, null, null);
				} else if (fontName.equals("Arial")) {
					insertFontDescription(doc4Fonts, "Arial", null, "Arial", "swiss", "variable", null);
				} else if (fontName.equals("Book Antiqua")) {
					insertFontDescription(doc4Fonts, "Book Antiqua", null, "Book Antiqua", "roman", "variable", null);
				} else if (fontName.equals("Calibri")) {
					insertFontDescription(doc4Fonts, "Calibri", null, "Calibri", "swiss", "variable", "2 15 5 2 2 2 4 3 2 4");
				} else if (fontName.equals("Cambria")) {
					insertFontDescription(doc4Fonts, "Cambria", null, "Cambria", "roman", "variable", "2 4 5 3 5 4 6 3 2 4");
				} else if (fontName.equals("Consolas")) {
					insertFontDescription(doc4Fonts, "Consolas", null, "Consolas", "modern", "fixed", null);
				} else if (fontName.equals("Courier New")) {
					insertFontDescription(doc4Fonts, "Courier New", null, "Courier New", "modern", "fixed", null);
				} else if (fontName.equals("Courier")) {
					insertFontDescription(doc4Fonts, "Courier", null, "Courier", "modern", "fixed", null);
				} else if (fontName.equals("Georgia")) {
					insertFontDescription(doc4Fonts, "Georgia", null, "Georgia", "roman", "variable", null);
				} else if (fontName.equals("Helvetica")) {
					insertFontDescription(doc4Fonts, "Helvetica", null, "Helvetica", "swiss", null, null);
				} else if (fontName.equals("Impact")) {
					insertFontDescription(doc4Fonts, "Impact", null, "Impact", "swiss", "variable", null);
				} else if (fontName.equals("Mangal")) {
					insertFontDescription(doc4Fonts, "Mangal", null, "Mangal", "system", "variable", null);
				} else if (fontName.equals("Mangal1")) {
					insertFontDescription(doc4Fonts, "Mangal1", null, "Mangal", null, null, null);
				} else if (fontName.equals("Microsoft YaHei")) {
					insertFontDescription(doc4Fonts, "Microsoft YaHei", null, "Microsoft YaHei", "system", "variable", null);
				} else if (fontName.equals("MS Gothic")) {
					insertFontDescription(doc4Fonts, "MS Gothic", null, "MS Gothic", "modern", "fixed", "2 11 6 9 7 2 5 8 2 4");
				} else if (fontName.equals("MS Mincho")) {
					insertFontDescription(doc4Fonts, "MS Mincho", null, "MS Mincho", "modern", "fixed", "2 2 6 9 4 2 5 8 3 4");
				} else if (fontName.equals("Palatino")) {
					insertFontDescription(doc4Fonts, "Palatino", null, "Palatino", "roman", null, null);
				} else if (fontName.equals("SimSun")) {
					insertFontDescription(doc4Fonts, "SimSun", null, "SimSun", "system", "variable", null);
				} else if (fontName.equals("Tahoma")) {
					insertFontDescription(doc4Fonts, "Tahoma", null, "Tahoma", "swiss", "variable", null);
				} else if (fontName.equals("Times New Roman")) {
					insertFontDescription(doc4Fonts, "Times New Roman", null, "Times New Roman", "roman", "variable", "2 2 6 3 5 4 5 2 3 4");
				} else if (fontName.equals("Times")) {
					insertFontDescription(doc4Fonts, "Times", null, "Times", "roman", null, null);
				} else if (fontName.equals("Verdana")) {
					insertFontDescription(doc4Fonts, "Verdana", null, "Verdana", "swiss", "variable", null);
				}
			}
		}
	}

	static private void mapParagraphProperties(JSONObject attrs, StyleParagraphPropertiesElement propertiesElement) {

		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				addBorderProperties(key, attrs, propertiesElement);
				addPaddingProperties(key, attrs, propertiesElement);
				addMarginProperties(key, attrs, propertiesElement);
				// fo:background-color
				/*
				 <attribute name="fo:background-color">
				 <choice>
				 <value>transparent</value>
				 <ref name="color"/>
				 </choice>
				 </attribute>
				 <define name="color">
				 <data type="string">
				 <param name="pattern">#[0-9a-fA-F]{6}</param>
				 </data>
				 </define>			
				 */
				if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else /* 
				 The defined values for the fo:line-height attribute are:
				 * a value of type nonNegativeLength 
				 * normal: disables the effects of style:line-height-at-least and style:line-spacing.  
				 * a value of type percent  */ //				<attribute name="fo:line-height">
				//					<choice>
				//						<value>normal</value>
				//						<ref name="nonNegativeLength"/>
				//						<ref name="percent"/>
				//					</choice>
				//				</attribute>
				// { type: 'percent', value: 100 }																
				if (key.equals("lineHeight")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "line-height");
						} else {
							JSONObject lineHeight = (JSONObject) value;
							String lineHeightValue = getLineHeight(lineHeight);
							if (lineHeightValue != null && !lineHeightValue.isEmpty()) {
								propertiesElement.setFoLineHeightAttribute(lineHeightValue);
							}

						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} // see. https://dev-wiki-archiv.open-xchange.com/cgi-bin/twiki/view/Main/Operations#Paragraph_Formatting_Attributes
				// One of 'left', 'center', 'right', or 'justify'.
				// start, end, left, right, center or justify.
				else if (key.equals("alignment")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "text-align");
						} else {
							propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:text-align", (String) value);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} // see. https://dev-wiki-archiv.open-xchange.com/cgi-bin/twiki/view/Main/Operations#Paragraph_Formatting_Attributes
				//<attribute name="fo:text-indent">
				//	<choice>
				//		<ref name="length"/>
				//		<ref name="percent"/>
				//	</choice>
				//</attribute>
				else if (key.equals("indentFirstLine")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "text-indent");
						} else {
							Integer indent = (Integer) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:text-indent", ((indent / 100.0) + "mm"));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
//				
//	/**
//	 * see
//	 * https://dev-wiki-archiv.open-xchange.com/cgi-bin/twiki/view/Main/Operations#LineHeight
//	 */
//	JSONObject createLineHeightMap(String lineHeightValue) {
//		JSONObject lineHeight = new JSONObject();
//		try {
//			if (lineHeightValue.equals("normal")) {
//				lineHeight.put("type", "percent");
//				lineHeight.put("value", "100");
//			} else if (lineHeightValue.contains("%")) {
//				lineHeight.put("type", "percent");
//				lineHeightValue = lineHeightValue.subSequence(0, lineHeightValue.indexOf('%')).toString();
//				lineHeight.put("value", lineHeightValue);
//			} else {
//				lineHeight.put("type", "fixed");
//				lineHeight.put("value", normalizeLength(lineHeightValue));
//			}
//		} catch (JSONException ex) {
//			Logger.getLogger(JsonOperationProducer.class.getName()).log(Level.SEVERE, null, ex);
//		}
//		return lineHeight;
//	}				
//				
//						else if (propName.equals("fo:line-height")) {
//							propValue = odfProps.get("fo:line-height");
//							JSONObject lineHeight = createLineHeightMap(propValue);
//							newProps.put("lineHeight", lineHeight);
//						}				
			}
		}
	}
	/* 
	 * 
	 } else if (propName.equals("style:width")) {
	 propValue = odfProps.get("style:width");
	 newProps.put("width", normalizeLength(propValue));
	 } else if (propName.equals("style:writing-mode")) {
	 propValue = odfProps.get("style:writing-mode");
	 newProps.put("writingMode", propValue);
	

	 */

	static private void mapTableProperties(JSONObject attrs, StyleTablePropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				// no padding, no border
				addMarginProperties(key, attrs, propertiesElement);
				if (key.equals("width")) {
					Object value = null;
					try {
						value = attrs.get(key);
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}

					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "width");
					} else {
						String width = (String) value;
						if (width != null && !width.equals(AUTO)) {
							propertiesElement.setStyleWidthAttribute(Integer.parseInt(width) / 1000.0 + "cm");
						}
					}
				} else if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}

	static private void mapPageProperties(JSONObject attrs, StylePageLayoutPropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				addBorderProperties(key, attrs, propertiesElement);
				addPaddingProperties(key, attrs, propertiesElement);
				addMarginProperties(key, attrs, propertiesElement);
				if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}

				}
			}
		}
	}

	static private void mapRowProperties(JSONObject attrs, StyleTableRowPropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}

				} else if (key.equals("height")) {
					try {
						Object value = attrs.get(key);
						// API ToDo: Only the minimum row height is being round-tripped. row height is being mapped to minimum row height!
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "min-row-height");
						} else {
							Integer rowHeight = (Integer) value;
							propertiesElement.setStyleMinRowHeightAttribute(((rowHeight / 100.0) + "mm"));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}

	static private void mapCellProperties(JSONObject attrs, StyleTableCellPropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				// No margin in ODF
				addBorderProperties(key, attrs, propertiesElement);
				addPaddingProperties(key, attrs, propertiesElement);
				if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}

	static private void mapColumnProperties(JSONObject attrs, StyleTableColumnPropertiesElement propertiesElement) {
	}

	static private void mapListProperties(JSONObject attrs, StyleListLevelPropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				if (key.equals("fontName")) {
					try {
						Object value = attrs.get(key);
						// == null does not work here..
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-name");
						} else {
							String fontName = (String) value;
							propertiesElement.setStyleFontNameAttribute(fontName);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontNameAsian")) {
					try {
						Object value = attrs.get(key);
						// == null does not work here..
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-name-asian");
						} else {
							String fontName = (String) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-name-asian", fontName);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("fontNameComplex")) {
					try {
						Object value = attrs.get(key);
						// == null does not work here..
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "font-name-complex");
						} else {
							String fontName = (String) value;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "style:font-name-complex", fontName);
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}

	static private void mapSectionProperties(JSONObject attrs, StyleSectionPropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				// No margin in ODF
				addBorderProperties(key, attrs, propertiesElement);
				addPaddingProperties(key, attrs, propertiesElement);
				if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}

	static private void mapGraphicProperties(JSONObject attrs, StyleGraphicPropertiesElement propertiesElement) {
		if (attrs != null) {
			boolean isMirroredHorizontalRemoved = false;
			boolean isMirroredHorizontal = false;
			boolean isMirroredVerticalRemoved = false;
			boolean isMirroredVertical = false;
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				addBorderProperties(key, attrs, propertiesElement);
				addPaddingProperties(key, attrs, propertiesElement);
				addMarginProperties(key, attrs, propertiesElement);
				if (key.equals("flipH")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							isMirroredHorizontalRemoved = true;
						} else if ((Boolean) value) {
							isMirroredHorizontal = true;
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("flipV")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							isMirroredVerticalRemoved = true;
						} else if ((Boolean) value) {
							isMirroredVertical = true;
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else /* http://docs.oasis-open.org/office/v1.2/os/OpenDocument-v1.2-os-part1.html#property-style_wrap
				 The style:wrap attribute specifies how text is displayed around a frame or graphic object.
				 The defined values for the style:wrap attribute are:
				 * biggest: text may wrap around the shape where the difference to the left or right page or column border is largest. 
				 * Mode=square Side=largest
				 * dynamic: text may wrap around both sides of the shape. The space for wrapping is set by the style:wrap-dynamic-threshold attribute. 20.393 
				 * UNSUPPORTED
				 * left: text wraps around the left side of the shape. 
				 * Mode=square	Side=left
				 * none: text does not wrap around the shape. 
				 * Mode=topAndBottom 
				 * parallel: text wraps around both sides of the shape. 
				 * Mode=square	Side=both	 
				 * right: text wraps around the right side of the shape. 
				 * Mode=square	Side=right
				 * run-through: text runs through the shape.
				 * Mode=through	Side=both   */ /* 
				 OX API: textWrapMode	One of 'none', 'square', 'tight', 'through', or 'topAndBottom'. (IMHO 'none' == 'topAndBottom', but the latter is implemented)
				 OX API: textWrapSide	Sides where text wraps around the image (only used if textWrapMode is set to 'square', 'tight', or 'through') (1). One of ['both', 'left', 'right', 'largest'	*/ if (key.equals("textWrapMode")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "wrap");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "wrap-contour");
						} else {
							String textWrapMode = (String) value;
							if (textWrapMode.equals("topAndBottom")) {
								propertiesElement.setStyleWrapAttribute("none");
								propertiesElement.setStyleWrapContourAttribute(Boolean.FALSE);
							} else {
								String textWrapSide = (String) attrs.optString("textWrapSide");
								if (textWrapSide != null) {
									if (textWrapMode.equals("square")) {
										if (textWrapSide.equals("largest")) {
											propertiesElement.setStyleWrapAttribute("biggest");
										} else if (textWrapSide.equals("left")) {
											propertiesElement.setStyleWrapAttribute("left");
										} else if (textWrapSide.equals("both")) {
											propertiesElement.setStyleWrapAttribute("parallel");
										} else if (textWrapSide.equals("right")) {
											propertiesElement.setStyleWrapAttribute("right");
										}
										propertiesElement.setStyleWrapContourAttribute(Boolean.FALSE);
									} else if (textWrapMode.equals("through")) {
										propertiesElement.setStyleWrapAttribute("run-through");
										propertiesElement.setStyleWrapContourAttribute(Boolean.FALSE);
									}
								}
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else /* @style:horizontal-pos:
				 The defined values for the style:horizontal-pos attribute are:
				 * center: horizontal alignment of a frame should be centered relative to the specified area. 
				 * anchorHorAlign=center
				 * from-inside: on pages with an odd page number the left edge of the specific area is taken as the horizontal alignment of a frame. On pages with an even page number the right edge of the specified area is taken. Attribute svg:x associated with the frame element specifies the horizontal position of the frame from the edge which is taken.  
				 * UNSUPPORTED
				 * from-left: the svg:x attribute associated with the frame element specifies the horizontal position of the frame from the left edge of the specified area. 
				 * anchorHorAlign=offset 
				 * inside: on pages with an odd page number the horizontal alignment of a frame is the same as for the attribute value left. On pages with an even page number the horizontal alignment of a frame is the same as for the attribute value right.  
				 * anchorHorAlign=inside
				 * left: horizontal alignment of a frame should be left aligned relative to the specified area. 
				 * anchorHorAlign=left
				 * outside: on pages with an odd page number the horizontal alignment of a frame is the same as for the attribute value right. On pages with an even page number the horizontal alignment of a frame is the same as for the attribute value left.  
				 * anchorHorAlign=outside 
				 * right: horizontal alignment of a frame should be right aligned relative to the specified area. 
				 * anchorHorAlign=right
				 If the attribute value is not from-left and not from-inside, the svg:x attribute associated with the frame element is ignored for text documents.						 */ // OX API:
				// anchorHorAlign: Horizontal anchor position:	One of 'left', 'right', 'center', 'inside', 'outside', or 'offset'.   	
				// anchorHorOffset: Horizontal position offset (only used if anchorHorAlign is set to 'offset') 
				if (key.equals("anchorHorAlign")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "horizontal-pos");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "horizontal-rel");
						} else {
							String anchorHorAlign = (String) value;
							// This is the default in our constellation
							propertiesElement.setStyleHorizontalRelAttribute("paragraph");

							if (anchorHorAlign.equals("center")) {
								propertiesElement.setStyleHorizontalPosAttribute("center");
							} else if (anchorHorAlign.equals("offset")) {
								propertiesElement.setStyleHorizontalPosAttribute("from-left");
							} else if (anchorHorAlign.equals("left")) {
								propertiesElement.setStyleHorizontalPosAttribute("left");
							} else if (anchorHorAlign.equals("right")) {
								propertiesElement.setStyleHorizontalPosAttribute("right");
							} else if (anchorHorAlign.equals("inside")) {
								propertiesElement.setStyleHorizontalPosAttribute("inside");
							} else if (anchorHorAlign.equals("outside")) {
								propertiesElement.setStyleHorizontalPosAttribute("outside");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("anchorHorOffset")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.SVG.getUri(), "x");
						} else {
							Integer x = (Integer) value;
							// This is the default in our constellation
							propertiesElement.setSvgXAttribute(x / 100.0 + "mm");
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("anchorVertAlign")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "vertical-pos");
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "vertical-rel");
						} else {
							String anchorVertAlign = (String) value;
							// This is the default in our constellation
							propertiesElement.setStyleVerticalRelAttribute("paragraph");

							if (anchorVertAlign.equals("center")) {
								propertiesElement.setStyleVerticalPosAttribute("center");
							} else if (anchorVertAlign.equals("offset")) {
								propertiesElement.setStyleVerticalPosAttribute("from-top");
							} else if (anchorVertAlign.equals("bottom")) {
								propertiesElement.setStyleVerticalPosAttribute("bottom");
							} else if (anchorVertAlign.equals("top")) {
								propertiesElement.setStyleVerticalPosAttribute("top");
							} else if (anchorVertAlign.equals("inside")) {
								propertiesElement.setStyleVerticalPosAttribute("inside");
							} else if (anchorVertAlign.equals("outside")) {
								propertiesElement.setStyleVerticalPosAttribute("outside");
							}
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				} else if (key.equals("anchorVertOffset")) {
					try {
						Object value = attrs.get(key);
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.SVG.getUri(), "y");
						} else {
							Integer y = (Integer) value;
							// This is the default in our constellation
							propertiesElement.setSvgYAttribute(y / 100.0 + "mm");
						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}


			//					} else if (propName.contains("svg:x")) {
			//						int x = normalizeLength(odfProps.get("svg:x"));
			//						if(x != 0){
			//							newProps.put("anchorHorOffset", x);
			//						}


			// Two attributes are evaluated via boolean flags:
			if (isMirroredHorizontalRemoved && isMirroredVerticalRemoved) {
				propertiesElement.removeAttributeNS(OdfDocumentNamespace.STYLE.getUri(), "mirror");
			} else if (isMirroredHorizontal && isMirroredVertical) {
				propertiesElement.setStyleMirrorAttribute("horizontal vertical");
			} else if (isMirroredVertical) {
				propertiesElement.setStyleMirrorAttribute("vertical");
			} else if (isMirroredHorizontal) {
				propertiesElement.setStyleMirrorAttribute("horizontal");
			}
		}
	}

	static private void mapChartProperties(JSONObject attrs, StyleChartPropertiesElement propertiesElement) {
	}

	static private void mapDrawingProperties(JSONObject attrs, StyleDrawingPagePropertiesElement propertiesElement) {
	}

	static private void mapRubyProperties(JSONObject attrs, StyleRubyPropertiesElement propertiesElement) {
	}

	static private void mapHeaderFooterProperties(JSONObject attrs, StyleHeaderFooterPropertiesElement propertiesElement) {
		if (attrs != null) {
			Set<String> propKeys = attrs.keySet();
			for (String key : propKeys) {
				// No border, padding in ODF
				addMarginProperties(key, attrs, propertiesElement);
				if (key.equals("fillColor")) {
					try {
						Object value = attrs.get(key);
						// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
						if (value == null || value.equals(null)) {
							propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "background-color");
						} else {
							JSONObject color = (JSONObject) value;
							propertiesElement.setFoBackgroundColorAttribute(getColor(color, TRANSPARENT));


						}
					} catch (JSONException ex) {
						Logger.getLogger(JsonOperationConsumer.class
								.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}
		}
	}

	//	fo:border="0.5pt solid #000000" 
	// {\"width\":2,
	//  \"style\":\"solid\"
	//  \"color\":{\"value\":\"000000\",\"type\":\"rgb\"}}		
	static private String getBorder(JSONObject border) {
		String style = null;
		String borderValue = "";
		if (border.hasAndNotNull("style")) {
			//  'none', 'single', 'double', 'dotted', 'dashed', 'outset', or 'inset'
			style = border.optString("style");
			if (style.equals("none")) {
				borderValue = "none";
			} else {
				if (border.hasAndNotNull("width")) {
					double width = border.optInt("width") / 100.0;
					borderValue = width + "mm ";
				} else {
					// DEFAULT BORDER: 0.002cm solid #000000" 
					borderValue = "0.02mm ";
				}
				if (border.hasAndNotNull("style")) {
					//  'none', 'single', 'double', 'dotted', 'dashed', 'outset', or 'inset'
					style = border.optString("style");
					if (style.equals("single")) {
						borderValue = borderValue.concat("solid ");
					} else {
						borderValue = borderValue.concat(style + " ");
					}
				} else {
					// DEFAULT BORDER: 0.002cm solid #000000" 
					borderValue = borderValue.concat("solid ");
				}
				if (border.hasAndNotNull("color")) {
					JSONObject color = border.optJSONObject("color");
					borderValue = borderValue.concat(getColor(color, BLACK));
				} else {
					// DEFAULT BORDER: 0.002cm solid #000000" 
					borderValue = borderValue.concat("#000000");
				}
			}
		}
		return borderValue;
	}

	/**
	 * Supporting "normal", "percentage" and a none-negative number
	 */
	private static String getLineHeight(JSONObject lineHeight) {
		String lineHeightValue = null;
		try {
			String type = lineHeight.getString("type");
			if (type.equals("percent")) {
				String value = lineHeight.getString("value");
				if (value != null && value.equals("100")) {
					lineHeightValue = "normal";
				} else if (value != null && !value.isEmpty()) {
					lineHeightValue = value.concat("%");
				}
			} else if (type.equals("fixed")) {
				double value = lineHeight.optInt("value") / 100.0;
				lineHeightValue = value + "mm";


			}
		} catch (JSONException ex) {
			Logger.getLogger(JsonOperationConsumer.class
					.getName()).log(Level.SEVERE, null, ex);
		}
		return lineHeightValue;
	}

	private static String getColor(JSONObject color, String autoColor) {
		String colorValue = "";
		try {
			if (color != null) {
				String type = color.getString("type");
				if (type.equals(AUTO)) {
					colorValue = autoColor;
				} else if (type.equals("rgb")) {
					colorValue = '#' + color.getString("value");
				} else {
					return '#' + color.getString("fallbackValue");


				}
			}
// {"color":{"type":"scheme","value":"accent1","transformations":[{"type":"shade","value":74902}],"fallbackValue":"376092"}			
		} catch (JSONException ex) {
			Logger.getLogger(JsonOperationConsumer.class
					.getName()).log(Level.SEVERE, null, ex);
		}
		return colorValue;
	}

	static private void addBorderProperties(String key, JSONObject attrs, OdfStylePropertiesBase propertiesElement) {
		if (key.contains("border")) {
//			if (key.equals("border")) {
//				try {
//					Object value = attrs.get(key);
//					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
//					if (value == null || value.equals(null)) {
//						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "border");
//					} else {
//						// {\"width\":2,
//						// \"style\":\"solid\"
//						// \"color\":{\"value\":\"000000\",\"type\":\"rgb\"}}							
//						JSONObject border = (JSONObject) value;
//						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(),"border", getBorder(border));
//					}
//				} catch (JSONException ex) {
//					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
//				}
//			} else 
			if (key.equals("borderLeft")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "border-left");
					} else {
						// {\"width\":2,
						// \"style\":\"solid\"
						// \"color\":{\"value\":\"000000\",\"type\":\"rgb\"}}							
						JSONObject border = (JSONObject) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:border-left", getBorder(border));
						// oppose to the other border styles, the padding was added to border
						if (border.hasAndNotNull("space")) {
							double padding = border.optInt("space") / 100.0;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-left", padding + "mm");
						}
					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("borderRight")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "border-right");
					} else {
						// {\"width\":15,
						// \"style\":\"solid\"
						// \"color\":{\"value\":\"000000\",\"type\":\"rgb\"}} / // \"color\":{\"type\":\"auto\"}}						
						// \"space\":\140
						JSONObject border = (JSONObject) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:border-right", getBorder(border));
						// oppose to the other border styles, the padding was added to border
						if (border.hasAndNotNull("space")) {
							double padding = border.optInt("space") / 100.0;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-right", padding + "mm");
						}
					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("borderTop")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "border-top");
					} else {
						// {\"width\":2,
						// \"style\":\"solid\"
						// \"color\":{\"value\":\"000000\",\"type\":\"rgb\"}}							
						JSONObject border = (JSONObject) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:border-top", getBorder(border));
						// oppose to the other border styles, the padding was added to border
						if (border.hasAndNotNull("space")) {
							double padding = border.optInt("space") / 100.0;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-top", padding + "mm");
						}
					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("borderBottom")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "border-bottom");
					} else {
						// {\"width\":2,
						// \"style\":\"solid\"
						// \"color\":{\"value\":\"000000\",\"type\":\"rgb\"}}							
						JSONObject border = (JSONObject) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:border-bottom", getBorder(border));
						// oppose to the other border styles, the padding was added to border
						if (border.hasAndNotNull("space")) {
							double padding = border.optInt("space") / 100.0;
							propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-bottom", padding + "mm");
						}
					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			}
		}
	}

	static private void addMarginProperties(String key, JSONObject attrs, OdfStylePropertiesBase propertiesElement) {
		if (key.contains("margin") || key.contains("indent")) {
//			if (key.equals("margin")) {
//				try {
//					Object value = attrs.get(key);
//					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
//					if (value == null || value.equals(null)) {
//						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "margin");
//					} else {
//						Integer width = (Integer) value;
//						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(),"margin", ((width / 100.0) + "mm"));
//					}
//				} catch (JSONException ex) {
//					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
//				}
//			} else 
			if (key.equals("marginBottom")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "margin-bottom");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:margin-bottom", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("marginLeft") || key.equals("indentLeft")) {
//FIX API   			} else if (key.equals("indentLeft")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "margin-left");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:margin-left", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("marginRight") || key.equals("indentRight")) {
//FIX API   			} else if (key.equals("indentRight")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "margin-right");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:margin-right", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("marginTop")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "margin-top");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:margin-top", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			}
		}
	}

	static private void addPaddingProperties(String key, JSONObject attrs, OdfStylePropertiesBase propertiesElement) {
		if (key.contains("padding")) {
//				try {
//					Object value = attrs.get(key);
//					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
//					if (value == null || value.equals(null)) {
//						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "padding");
//					} else {
//						Integer width = (Integer) value;
//						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(),"padding", ((width / 100.0) + "mm"));
//					}
//				} catch (JSONException ex) {
//					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
//				}
////			} 
//			else if (key.equals("padding")) {
//				try {
//					Object value = attrs.get(key);
//					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
//					if (value == null || value.equals(null)) {
//						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "padding");
//					} else {
//						Integer width = (Integer) value;
//						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(),"padding", ((width / 100.0) + "mm"));
//					}
//				} catch (JSONException ex) {
//					Logger.getLogger(JsonOperationConsumer.class.getName()).log(Level.SEVERE, null, ex);
//				}
//			} else 
			if (key.equals("paddingBottom")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "padding-bottom");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-bottom", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("paddingLeft")) {
				try {
					Object value = attrs.get(key);
					// ToDo: Test with latest JSON Library and adapt accordingly for all occurences
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "padding-left");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-left", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("paddingRight")) {
				try {
					Object value = attrs.get(key);
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "padding-right");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-right", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			} else if (key.equals("paddingTop")) {
				try {
					Object value = attrs.get(key);
					if (value == null || value.equals(null)) {
						propertiesElement.removeAttributeNS(OdfDocumentNamespace.FO.getUri(), "padding-top");
					} else {
						Integer width = (Integer) value;
						propertiesElement.setAttributeNS(OdfDocumentNamespace.FO.getUri(), "fo:padding-top", ((width / 100.0) + "mm"));


					}
				} catch (JSONException ex) {
					Logger.getLogger(JsonOperationConsumer.class
							.getName()).log(Level.SEVERE, null, ex);
				}
			}
		}
	}
}
//					} else if (propName.equals("style:width")) {
//						propValue = odfProps.get("style:width");
//						newProps.put("width", normalizeLength(propValue));
//					} else if (propName.equals("style:writing-mode")) {
//						propValue = odfProps.get("style:writing-mode");
//						newProps.put("writingMode", propValue);
//			}
//			} else if (styleFamilyGroup.equals("page")) {
//				for (String propName : odfProps.keySet()) {
//					try {
//						if (propName.equals("fo:page-width")) {
//							propValue = odfProps.get("fo:page-width");
//							newProps.put("width", normalizeLength(propValue));
//						} else if (propName.equals("fo:page-height")) {
//							propValue = odfProps.get("fo:page-height");
//							newProps.put("height", normalizeLength(propValue));
//						} else if (propName.equals("style:print-orientation")) {
//							propValue = odfProps.get("style:print-orientation");
//							newProps.put("printOrientation", propValue);
//						} else if (propName.equals("style:num-format")) {
//							propValue = odfProps.get("style:num-format");
//							newProps.put("numberFormat", propValue);
//						}
//					} catch (JSONException ex) {
//						Logger.getLogger(JsonOperationProducer.class.getName()).log(Level.SEVERE, null, ex);
//					}
//				}
//				}
//			}
//	
//	private static Node convertXMLToNode(OdfDocument doc, String xmlString) throws Exception{
//		
//		try {
//	
//				XMLReader xmlReader = doc.getPackage().getXMLReader();
//				xmlReader.setContentHandler(handler);
//				InputSource xmlSource = new InputSource(fileStream);
//				xmlReader.parse(xmlSource);
//			}
//			
//		} catch (Exception ex) {
//			Logger.getLogger(OdfPackageDocument.class.getName()).log(Level.SEVERE, null, ex);
//		} 
////		return  DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new ByteArrayInputStream(xmlString.getBytes())).getDocumentElement();
//	}

